# Standardization of data types
import abc
import os
import re
from abc import ABC
import pandas as pd
import numpy as np
from typing import Any, Union, get_origin, get_args


class Type(ABC):
    """Abstract base class for all supported data types"""

    pass


class Boolean(Type):
    """Abstract base class for all boolean types; `bool` and `np.bool_` are
    subclasses"""

    pass


Boolean.register(bool)
Boolean.register(np.bool_)


class Toggle(Type):
    """
    Convenience class for handling toggle variables, which are either off or on.
    """

    on_values = [True, 1, "1", "ON", "On", "on"]
    off_values = [False, 0, "0", "OFF", "Off", "off"]

    def __init__(self, state: Union[str, bool, int, type]):
        if hasattr(state, "on"):
            self.on = state.on
        elif state in self.on_values:
            self.on = True
        elif state in self.off_values:
            self.on = False
        else:
            raise ValueError(f"toggle was initialized with invalid state {state}")

    def __bool__(self):
        return True if self.on else False

    def __int__(self):
        return 1 if self.on else 0

    def __str__(self):
        return "ON" if self.on else "OFF"

    def __eq__(self, other):
        if hasattr(other, "on"):
            return self.on == other.on
        else:
            if self.on and other in self.on_values:
                return True
            elif not self.on and other in self.off_values:
                return True
            else:
                return False


ON = Toggle("ON")
OFF = Toggle("OFF")


class Integer(Type):
    """Abstract base class for all integer types; `int` and `np.integer` are
    subclasses"""

    pass


Integer.register(int)
Integer.register(np.integer)


class Float(Type):
    """Abstract base class for all float types; `float` and `np.floating` are
    subclasses"""

    pass


Float.register(float)
Float.register(np.floating)


class String(Type):
    """Abstract base class for all string types; `str` and `np.str_` are
    subclasses"""

    pass


String.register(str)
String.register(np.str_)


class Array(Type):
    """
    Abstract base class for all array-like types, essentially any commonly
    used type that can be indexed; `list`, `tuple`, `numpy.ndarray`,
    `pandas.Series` and `pandas.Dataframe` are subclasses
    """

    pass


Array.register(list)
Array.register(tuple)
Array.register(np.ndarray)
Array.register(pd.Series)
Array.register(pd.DataFrame)


supported = {
    key: value
    for key, value in vars().items()
    if type(value) is abc.ABCMeta and issubclass(value, Type)
}


def recast(value: Any, to: type = Type) -> Union[Type, None]:
    """
    Convert a value into the appropriate type for the information it contains.

    Booleans are converted into numpy booleans; integers are converted into
    64-bit numpy integers; floats are converted into 64-bit numpy floats.

    Array-like values are converted into the analogous numpy array.

    Strings are inspected to determine if they represent boolean or numerical
    values and, if so, recasts values to the appropriate types. If a string
    is a path in either the working directory or the parent directory thereof,
    it is converted into the full absolute path. If a string does not
    represent boolean or numerical values and is not a path, then this function
    just returns the same string.

    If the value argument does not fit into one of the above categories, a
    warning will be printed and None will be returned.

    :param value: (Any) the value whose type needs converting
    :param to: (Type) optional keyword argument indicating which type to
                       convert to; default value is `Type` which indicates
                       that the type should be inferred based on the value.
    """

    if to != Type:
        if value is None:
            return None

        for dtype in np.array([to], dtype=object).flatten():
            # Recast to desired type
            try:

                if get_origin(dtype) is Union:
                    # Expand type unions
                    return recast(value, to=get_args(dtype))

                if issubclass(dtype, Boolean):
                    return np.bool_(value)
                elif issubclass(dtype, Toggle):
                    return Toggle(value)
                elif issubclass(dtype, Integer):
                    return np.int64(value)
                elif issubclass(dtype, Float):
                    return np.float64(value)
                elif issubclass(dtype, String):
                    return np.str_(value)
                elif issubclass(dtype, Array) and np.ndim(value) > 0:
                    return np.array(value)

            except ValueError:
                pass

        print(f"Warning: unable to recast value {value} to type {to}")

        return None

    else:
        # infer type
        if isinstance(value, Boolean):
            return np.bool_(value)
        elif isinstance(value, Toggle):
            return value
        elif isinstance(value, Integer):
            return np.int64(value)
        elif isinstance(value, Float):
            return np.float64(value)
        elif isinstance(value, String):
            if value.lower() == "true":
                return np.bool_(True)
            elif value.lower() == "false":
                return np.bool_(False)
            elif re.fullmatch("[0-9]+", value):  # integer
                return np.int64(value)
            elif re.fullmatch("[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?", value):
                # float
                return float(value)
            elif value in (Toggle.on_values + Toggle.off_values):
                return Toggle(value)
            elif os.path.isfile(value):  # path in the current working directory
                return os.path.abspath(value)
            elif os.path.isfile(os.path.join("..", value)):  # ... up one level
                return os.path.abspath(os.path.join("..", value))
            else:
                return value  # must be an actual string
        if isinstance(value, Array):  # value is an array
            np_array = np.array(value)  # convert to numpy array
            rep_elem = np_array.flatten()[0]  # representative element
            return np_array.astype(type(recast(rep_elem)))
        else:
            print(f"Warning: unable to recast value {value} of type {type(value)}")

            return None



# Tools for defining and running experiments
import collections
import datetime
import importlib
import os
import pathlib
import sys
import threading
import time
import tkinter as tk
from tkinter.filedialog import askopenfilename

import numpy as np
import pandas as pd
import pykwalify.errors
from pykwalify.core import Core as YamlValidator
from ruamel.yaml import YAML

from empyric import variables as _variables
from empyric import adapters as _adapters
from empyric import graphics as _graphics
from empyric import instruments as _instruments
from empyric import routines as _routines
from empyric.tools import convert_time, Clock
from empyric.types import recast, Boolean, Toggle, Integer, Float


class Experiment:
    """
    An iterable class which represents an experiment; iterates through any
    assigned routines,and retrieves and stores the values of all experiment
    variables.
    """

    # Possible statuses of an experiment
    READY = "Ready"  # Experiment is waiting to start
    RUNNING = "Running"  # Experiment is running
    HOLDING = "Holding"  # Routines are stopped, but measurements are ongoing
    STOPPED = "Stopped"  # Both routines and measurements are stopped
    TERMINATED = "Terminated"

    # Experiment has either finished or has been terminated by the user

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        prior_base_status = self._status.split(":")[0]
        new_base_status = status.split(":")[0]

        # Only allow change if the status is unlocked,
        # or if the base status is the same
        if not self.status_locked or new_base_status == prior_base_status:
            self._status = status

    @property
    def ready(self):
        return "Ready" in self.status

    @property
    def running(self):
        return "Running" in self.status

    @property
    def holding(self):
        return "Holding" in self.status

    @property
    def stopped(self):
        return "Stopped" in self.status

    @property
    def terminated(self):
        return "Terminated" in self.status

    def __init__(self, variables, routines=None, end=None):
        self.variables = variables
        # dict of the form {..., name: variable, ...}

        # Used to block evaluation of expressions
        # until their dependee variables are evaluated
        self.eval_events = {name: threading.Event() for name in variables}

        if routines:
            self.routines = routines
            # dictionary of the form {..., name: (variable_name, routine), ...}
        else:
            self.routines = {}

        if end:
            if end.lower() == "with routines":
                self.end = max([routine.end for routine in routines.values()])
            else:
                self.end = end
        else:
            self.end = np.inf

        self.clock = Clock()
        self.clock.start()

        self.timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

        self.data = pd.DataFrame(columns=["Time"] + list(variables.keys()))

        self.state = pd.Series({column: None for column in self.data.columns})
        self.state["Time"] = 0

        self._status = Experiment.READY
        self.status_locked = True
        # can only be unlocked by the start, hold, stop and terminate methods

        self.saved = []  # list of saved data entries

    def __next__(self):
        # Start the clock on first call
        if self.state.name is None:  # first step of the experiment
            self.start()
            self.status = Experiment.RUNNING + ": initializing..."

        # Update time
        self.state["Time"] = self.clock.time
        self.state.name = datetime.datetime.now()

        # If experiment is stopped, just return the knob & parameter values,
        # and nullify meter & expression values
        if self.stopped:
            threads = {}
            for name, variable in self.variables.items():
                is_knob = isinstance(variable, _variables.Knob)
                is_parameter = isinstance(variable, _variables.Parameter)

                if is_knob or is_parameter:
                    threads[name] = threading.Thread(
                        target=self._update_variable, args=(name,)
                    )
                    threads[name].start()
                else:
                    self.state[name] = None

            # Wait for all variable update threads to finish
            for name, thread in threads.items():
                self.status = Experiment.RUNNING + f": retrieving {name}"
                thread.join()

            return self.state

        # If the experiment is running, apply new settings to knobs
        # according to the routines (if there are any)
        if self.running:
            # Update each routine in its own thread
            threads = {}
            for name, routine in self.routines.items():
                threads[name] = threading.Thread(
                    target=self._update_routine, args=(name,)
                )
                threads[name].start()

            # Wait for all routine threads to finish
            for name, thread in threads.items():
                self.status = Experiment.RUNNING + f": executing {name}"
                thread.join()

            self.status = Experiment.RUNNING

        # Get all variable values if experiment is running or holding
        if self.running or self.holding:
            for event in self.eval_events.values():
                event.clear()

            # Run each measure / get operation in its own thread
            threads = {}
            for name in self.variables:
                threads[name] = threading.Thread(
                    target=self._update_variable, args=(name,)
                )
                threads[name].start()

            base_status = self.status

            # Wait for all threads to finish
            for name, thread in threads.items():
                self.status = base_status + f": retrieving {name}"
                thread.join()

            self.status = base_status

            # Append new state to experiment data set
            self.data.loc[self.state.name] = self.state

        # End the experiment, if the duration of the experiment has passed
        if self.clock.time > self.end:
            self.terminate()

        if self.terminated:
            raise StopIteration

        return self.state

    def __iter__(self):
        return self

    def _update_variable(self, name):
        """Retrieve and store a variable value"""

        try:
            if isinstance(self.variables[name], _variables.Expression):
                for dependee in self.variables[name].definitions:
                    event = self.eval_events[dependee]
                    event.wait()  # wait for dependee to be evaluated

            value = self.variables[name].value

            self.eval_events[name].set()
            # unblock threads evaluating dependents

            if np.size(value) > 1:  # store array data as CSV files
                if np.any(value):
                    # only save non-empty arrays
                    dataframe = pd.DataFrame(
                        {name: value}, index=[self.state.name] * len(value)
                    )
                    path = name.replace(" ", "_") + "_"
                    path += self.state.name.strftime("%Y%m%d-%H%M%S") + ".csv"
                    dataframe.to_csv(path)
                    self.state[name] = path
                else:
                    self.state[name] = None
            else:
                self.state[name] = value
        except BaseException as err:
            self.terminate()
            raise err

    def _update_routine(self, name):
        """Update a routine according to the current state"""

        try:
            self.routines[name].update(self.state)
        except BaseException as err:
            self.terminate()
            raise err

    def save(self, directory=None):
        """
        Save the experiment dataframe to a CSV file

        :param directory: (path) (optional) directory to save data to,
                          if different from working directory
        :return: None
        """

        base_status = self.status
        self.status = base_status + ": saving data"

        path = f"data_{self.timestamp}.csv"

        if directory:
            path = os.path.join(directory, path)

        if not os.path.exists(path):
            # if this is a new file, write column headers
            with open(path, "a") as data_file:
                data_file.write("," + ",".join(self.data.columns) + "\n")

        unsaved = np.setdiff1d(self.data.index, self.saved)

        with open(path, "a") as data_file:
            for line in unsaved:
                line_data = ",".join(map(str, list(self.data.loc[line])))
                data_file.write(str(line) + "," + line_data + "\n")

        self.saved = self.saved + list(unsaved)

        self.status = base_status

    def start(self):
        """
        Start the experiment: clock starts/resumes, routines resume,
        measurements continue

        :return: None
        """
        self.clock.start()

        self.status_locked = False
        self.status = Experiment.RUNNING
        self.status_locked = True

    def hold(self, reason=None):
        """
        Hold the experiment: clock stops, routines stop, measurements continue

        :return: None
        """
        self.clock.stop()

        self.status_locked = False
        self.status = Experiment.HOLDING
        if reason:
            self.status = self.status + ": " + reason
        self.status_locked = True

    def stop(self, reason=None):
        """
        Stop the experiment: clock stops, routines stop, measurements stop

        :return: None
        """
        self.clock.stop()

        self.status_locked = False
        self.status = Experiment.STOPPED
        if reason:
            self.status = self.status + ": " + reason
        self.status_locked = True

    def terminate(self, reason=None):
        """
        Terminate the experiment: clock, routines and measurements stop,
        data is saved and StopIteration is raised

        :return: None
        """
        self.stop()
        self.save()

        self.status_locked = False
        self.status = Experiment.TERMINATED
        if reason:
            self.status = self.status + ": " + reason
        self.status_locked = True

        # End routines
        for routine in self.routines.values():
            routine.terminate()

    def __repr__(self):
        return "Experiment"


class Alarm:
    """
    Triggers if a condition among variables is met and indicates the response
    protocol
    """

    def __init__(self, condition, variables, protocol=None):
        self.trigger_variable = _variables.Expression(
            expression=condition, definitions=variables
        )

        if protocol:
            self.protocol = protocol
        else:
            self.protocol = "none"

    @property
    def triggered(self):
        return self.trigger_variable.value is True

    def __repr__(self):
        return "Alarm"


class Manager:
    """
    Utility class which sets up and manages experiments, based on runcards.
    When initallized, it uses the given runcard to construct an experiment and
    run it in a separate thread. The Manager also handles alarms as specified
    by the runcard.
    """

    def __init__(self, runcard=None):
        """
        :param runcard: (dict/str) runcard as a dictionary or path pointing to
        a YAML file
        """

        if runcard:
            self.runcard = runcard
        else:
            # Have user locate runcard, if none given
            root = tk.Tk()
            root.withdraw()
            self.runcard = askopenfilename(
                parent=root,
                title="Select Runcard",
                filetypes=[("YAML files", "*.yaml")],
            )

            if self.runcard == "":
                raise ValueError("a valid runcard was not selected!")

        if isinstance(self.runcard, str):
            if os.path.exists(self.runcard):
                dirname = os.path.dirname(self.runcard)
                if dirname != "":
                    os.chdir(os.path.dirname(self.runcard))
                    # go to runcard directory to put data in same location

                yaml = YAML()
                with open(self.runcard, "rb") as runcard_file:
                    self.runcard = yaml.load(runcard_file)  # load the runcard
            else:
                raise FileNotFoundError(f'invalid runcard path "{self.runcard}"')
        elif isinstance(self.runcard, dict):
            pass
        else:
            raise TypeError(
                f"runcard given to Manager must be a path to a YAML file "
                f"or a dictionary, not {type(self.runcard)}!"
            )

        converted_runcard = convert_runcard(self.runcard)

        self.description = converted_runcard["Description"]
        self.settings = converted_runcard["Settings"]
        self.instruments = converted_runcard["Instruments"]
        self.alarms = converted_runcard["Alarms"]
        self.plotter = converted_runcard.get("Plotter", None)

        self.experiment = converted_runcard["Experiment"]

        # Unpack settings
        self.followup = self.settings.get("follow-up", None)
        self.step_interval = convert_time(self.settings.get("step interval", 0.1))
        self.save_interval = convert_time(self.settings.get("save interval", 60))
        self.plot_interval = convert_time(self.settings.get("plot interval", 0.1))
        self.end = convert_time(self.settings.get("end", np.inf))

        self.experiment.end = self.end

        self.last_save = 0

        self.awaiting_alarms = {}  # dictionary of alarms that are triggered

    def run(self, directory=None):
        """
        Run the experiment defined by the runcard. A GUI shows experiment
        status, while the experiment is run in a separate thread.

        :param directory: (path) (optional) directory in which to run the
                          experiment if different from the working directory
        :return: None
        """

        top_dir = os.getcwd()

        if directory:
            os.chdir(directory)

        # Create a new directory for data storage
        experiment_name = self.runcard["Description"].get("name", "Experiment")
        working_dir = experiment_name + "-" + self.experiment.timestamp
        os.mkdir(working_dir)
        os.chdir(working_dir)

        # Save executed runcard alongside data for record keeping
        yaml = YAML()

        timestamped_path = f"{experiment_name}_{self.experiment.timestamp}.yaml"
        with open(timestamped_path, "w") as runcard_file:
            yaml.dump(self.runcard, runcard_file)

        # Run experiment loop in separate thread
        experiment_thread = threading.Thread(target=self._run)
        experiment_thread.start()

        # Set up the GUI for user interaction
        self.gui = _graphics.ExperimentGUI(
            self.experiment,
            alarms=self.alarms,
            instruments=self.instruments,
            title=self.description.get("name", "Experiment"),
            plotter=self.plotter,
            save_interval=self.save_interval,
            plot_interval=self.plot_interval,
        )

        self.gui.run()

        experiment_thread.join()

        # Disconnect instruments
        for instrument in self.instruments.values():
            instrument.disconnect()

        os.chdir(top_dir)  # return to the parent directory

        # Cancel follow-up if experiment terminated by user
        if self.experiment.terminated and "user" in self.experiment.status:
            self.followup = None

        # Execute the follow-up experiment if there is one
        if self.followup:
            if "yaml" in self.followup:
                self.__init__(self.followup)
                self.run(directory=directory)
            elif "repeat" in self.followup:
                self.__init__(self.runcard)
                self.run(directory=directory)

    def _run(self):
        # Run in a separate thread

        for state in self.experiment:
            step_start = time.time()

            # Save experimental data periodically
            next_save = self.last_save + self.save_interval
            if self.experiment.clock.time >= next_save:
                save_thread = threading.Thread(target=self.experiment.save)
                save_thread.start()
                self.last_save = self.experiment.clock.time

            # Check if any alarms are triggered and handle them
            for name, alarm in self.alarms.items():
                if alarm.triggered:
                    # Add to dictionary of triggered alarms, if newly triggered
                    if name not in self.awaiting_alarms:
                        self.awaiting_alarms[name] = {
                            "time": self.experiment.data["Time"].iloc[-1],
                            "status": self.experiment.status,
                        }

                    if "none" in alarm.protocol:
                        # do nothing (GUI will indicate that alarm is triggered)
                        break
                    if "hold" in alarm.protocol:
                        # stop routines but keep measuring until alarm is clear
                        self.experiment.hold(reason=name)
                        break
                    elif "stop" in alarm.protocol:
                        # stop routines and measurements until alarm is clear
                        self.experiment.stop(reason=name)
                        break
                    elif "terminate" in alarm.protocol:
                        # terminate the experiment
                        self.experiment.terminate(reason=name)
                        break
                    elif "yaml" in alarm.protocol:
                        # terminate the experiment and run another
                        self.experiment.terminate(reason=name)
                        self.followup = alarm.protocol
                        break
                    elif "check" in alarm.protocol:
                        # stop routines and measurements until user resumes
                        self.experiment.stop(reason=name)
                        break
                    else:
                        # terminate the experiment
                        self.experiment.terminate(reason=name)
                        break

                elif name in self.awaiting_alarms:
                    # Alarm was previously triggered but is now clear

                    if "check" not in alarm.protocol:
                        # alarm is not waiting for user to check

                        info = self.awaiting_alarms.pop(name)
                        # remove from dictionary of triggered alarms
                        prior_status = info["status"]

                        if len(self.awaiting_alarms) == 0:
                            # there are no more triggered alarms
                            # return to state of experiment prior to trigger
                            if "Running" in prior_status:
                                self.experiment.start()
                            if "Holding" in prior_status:
                                self.experiment.hold(reason=name + " cleared")
                            if "Stopped" in prior_status:
                                self.experiment.stop(reason=name + " cleared")

            step_end = time.time()

            remaining_time = self.step_interval - (step_end - step_start)

            if remaining_time > 0:
                time.sleep(remaining_time)

    def __repr__(self):
        return "Manager"


class RuncardError(BaseException):
    pass


def validate_runcard(runcard):
    is_dict = isinstance(runcard, dict)
    is_ordereddict = isinstance(runcard, collections.OrderedDict)

    if is_dict or is_ordereddict:
        # Create temporary runcard YAML file for PyKwalify to validate
        yaml = YAML()

        runcard_path = f"tmp_runcard_{time.time()}.yaml"

        with open(runcard_path, "w") as runcard_file:
            yaml.dump(runcard, runcard_file)

        validate_runcard(runcard_path)

        # Wait until temporary file has write access, then delete
        while not os.access(runcard_path, os.W_OK):
            time.sleep(0.1)

        os.remove(runcard_path)

        return True

    elif type(runcard) is not str:
        raise ValueError("runcard must be either dict or str.")

    validator = YamlValidator(
        source_file=runcard,
        schema_files=[
            os.path.join(pathlib.Path(__file__).parent, "runcard_schema.yaml")
        ],
    )

    try:
        validator.validate(raise_exception=True)
    except pykwalify.errors.SchemaError as err:
        raise RuncardError(err)

    return True


def convert_runcard(runcard):
    """
    :param runcard: (dict/str) runcard dictionary or path string

    :return: (dict) converted runcard in the form described below.

    Converts the sections into the relevant objects:

    * The Descriptions and Settings sections are unchanged.
    * The Instruments section is converted into a dictionary of corresponding
      ``Instrument`` instances.
    * The Variables and Routines sections are combined into an ``Experiment``
      instance.
    * The Alarms section is  converted into a dictionary of corresponding
      ``Alarm`` objects.
    * The Plots section is converted into a corresponding ``Plotter`` instance.
    """

    # if runcard argument is a path to a YAML file, load into a dictionary
    if type(runcard) == str:
        yaml = YAML()
        with open(runcard, "rb") as runcard_file:
            runcard = yaml.load(runcard_file)

    # Validate runcard format and contents
    validate_runcard(runcard)

    converted_runcard = runcard.copy()

    # Load any custom components
    custom_routines = {}
    custom_instruments = {}

    if "custom.py" in os.listdir():
        sys.path.insert(1, os.getcwd())
        custom = importlib.import_module("custom")

        for name, thing in custom.__dict__.items():
            if type(thing) == type:
                if issubclass(thing, _routines.Routine):
                    custom_routines[name] = thing
                if issubclass(thing, _instruments.Instrument):
                    custom_instruments[name] = thing

    # Instruments section
    available_instruments = {**_instruments.supported, **custom_instruments}

    instruments = {}
    for name, specs in runcard.get("Instruments", {}).items():
        specs = specs.copy()
        _type = specs.pop("type")
        address = specs.get("address", None)

        # Grab any keyword arguments for the adapter
        adapter_kwargs = {}
        for kwarg in _adapters.kwargs:
            if kwarg.replace("_", " ") in specs:
                adapter_kwargs[kwarg] = specs.pop(kwarg.replace("_", " "))

        # Any remaining keywords are instrument presets
        presets = {
            key: recast(value) for key, value in specs.get("presets", {}).items()
        }
        postsets = {
            key: recast(value) for key, value in specs.get("postsets", {}).items()
        }

        instrument_class = available_instruments[_type]
        instruments[name] = instrument_class(
            address=address, presets=presets, postsets=postsets, **adapter_kwargs
        )
        instruments[name].name = name

    converted_runcard["Instruments"] = instruments

    # Variables section
    variables = {}
    for name, specs in runcard["Variables"].items():
        if "meter" in specs:
            instrument = converted_runcard["Instruments"][specs["instrument"]]
            gate = specs.get("gate", None)
            gate = variables[gate] if gate else None
            variables[name] = _variables.Meter(
                meter=specs["meter"], instrument=instrument, gate=gate
            )
        elif "knob" in specs:
            instrument = converted_runcard["Instruments"][specs["instrument"]]
            variables[name] = _variables.Knob(
                knob=specs["knob"],
                instrument=instrument,
                lower_limit=specs.get("lower limit", None),
                upper_limit=specs.get("upper limit", None),
            )
        elif "expression" in specs:
            expression = specs["expression"]
            definitions = {}

            for symbol, var_name in specs["definitions"].items():
                expression = expression.replace(symbol, var_name)

                try:
                    definitions[var_name] = variables[var_name]
                except KeyError as undefined:
                    raise KeyError(
                        f"variable {undefined} is not defined for expression "
                        f"'{name}'"
                    )

            variables[name] = _variables.Expression(
                expression=expression, definitions=definitions
            )
        elif "server" in specs:
            server = specs["server"]
            alias = specs.get("alias", name)
            protocol = specs.get("protocol", None)
            settable = specs.get("settable", False)

            variables[name] = _variables.Remote(
                server=server, alias=alias, protocol=protocol, settable=settable
            )

        elif "parameter" in specs:
            variables[name] = _variables.Parameter(parameter=specs["parameter"])

    # Routines section
    available_routines = {**_routines.supported, **custom_routines}

    routines = {}
    if "Routines" in runcard:
        for name, specs in runcard["Routines"].items():
            specs = specs.copy()  # avoids modifying the runcard
            _type = specs.pop("type")

            specs = {key.replace(" ", "_"): value for key, value in specs.items()}

            # Convert list of knobs into dictionary
            knobs = np.array([specs.get("knobs", [])]).flatten()
            if knobs.size > 0:
                for knob in knobs:
                    if knob not in variables:
                        raise KeyError(
                            f"knob {knob} specified for routine {name} "
                            "is not in Variables!"
                        )

                specs["knobs"] = {name: variables[name] for name in knobs}

            routines[name] = available_routines[_type](**specs)

    converted_runcard["Experiment"] = Experiment(variables, routines=routines)

    # Alarms section
    alarms = {}
    if "Alarms" in runcard:
        for name, specs in runcard["Alarms"].items():
            alarm_variables = specs.copy().get("variables", {})
            condition = specs.copy()["condition"]

            for variable in specs.get("variables", {}):
                if variable not in variables:
                    raise KeyError(
                        f"variable {variable} specified for alarm {name} "
                        f"is not in Variables!"
                    )

            alphabet = "abcdefghijklmnopqrstuvwxyz"
            for var_name, variable in variables.items():
                # Variables can be called by name in the condition
                if var_name in condition:
                    temp_name = "".join(
                        [
                            alphabet[  # pylint: disable=invalid-sequence-index
                                np.random.randint(0, len(alphabet))
                            ]
                            for i in range(3)
                        ]
                    )
                    while temp_name in alarm_variables:
                        # make sure temp_name is not repeated
                        temp_name = "".join(
                            [
                                alphabet[  # pylint: disable=invalid-sequence-index
                                    np.random.randint(0, len(alphabet))
                                ]
                                for i in range(3)
                            ]
                        )

                    alarm_variables.update({temp_name: variable})
                    condition = condition.replace(var_name, temp_name)

                # Otherwise, they are defined in the alarm_variables
                for symbol, alarm_variable_name in alarm_variables.items():
                    if alarm_variable_name == var_name:
                        alarm_variables[symbol] = variable

            alarms.update(
                {
                    name: Alarm(
                        condition, alarm_variables, protocol=specs.get("protocol", None)
                    )
                }
            )

    converted_runcard["Alarms"] = alarms

    # Plots section
    if "Plots" in runcard:
        converted_runcard["Plotter"] = _graphics.Plotter(
            converted_runcard["Experiment"].data, settings=runcard["Plots"]
        )
    else:
        converted_runcard["Plotter"] = None

    return converted_runcard



from empyric.adapters import *
from empyric.collection.instrument import *
from empyric.types import Toggle, Float, ON, OFF


class Keithley2260B(Instrument):
    """
    Keithley 2260B power supply, usually either 360 W or 720 W
    """

    name = "Keithley2260B"

    supported_adapters = ((Serial, {"baud_rate": 115200, "write_termination": "\r\n"}),)

    knobs = ("max voltage", "max current", "output")

    # no presets or postsets for this instrument

    meters = ("voltage", "current")

    @measurer
    def measure_current(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response))

        return float(self.query("MEAS:CURR?", validator=validator))

    @measurer
    def measure_voltage(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response))

        return float(self.query("MEAS:VOLT?", validator=validator))

    @setter
    def set_max_voltage(self, voltage: Float):
        self.write("VOLT %.4f" % voltage)

    @setter
    def set_max_current(self, current: Float):
        self.write("CURR %.4f" % current)

    @setter
    def set_output(self, output: Toggle):
        if output == ON:
            self.write("OUTP:STAT:IMM ON")
        elif output == OFF:
            self.write("OUTP:STAT:IMM OFF")

    @getter
    def get_max_current(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response))

        return float(self.query("CURR?", validator=validator))

    @getter
    def get_max_voltage(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response))

        return float(self.query("VOLT?", validator=validator))


class BK9183B(Instrument):
    """
    B&K Precision Model 9183B (35V & 6A / 70V & 3A) power supply
    """

    name = "BK9183B"

    supported_adapters = ((Serial, {"baud_rate": 57600}),)

    knobs = ("max voltage", "max current", "output")

    # no presets or postsets for this instrument

    meters = ("voltage", "current")

    @setter
    def set_output(self, output: Toggle):
        if output == ON:
            self.write("OUT ON")
        elif output == OFF:
            self.write("OUT OFF")

    @measurer
    def measure_current(self):
        # sometimes the first measurement is lagged
        return [float(self.query("MEAS:CURR?")) for i in range(3)][-1]

    @measurer
    def measure_voltage(self):
        # sometimes the first measurement is lagged
        return [float(self.query("MEAS:VOLT?")) for i in range(3)][-1]

    @setter
    def set_max_current(self, current):
        self.write("SOUR:CURR " + str(current))

    @setter
    def set_max_voltage(self, voltage):
        self.write("SOUR:VOLT " + str(voltage))

    @getter
    def get_max_current(self):
        return float(self.query("SOUR:CURR?"))

    @getter
    def get_max_voltage(self):
        return float(self.query("SOUR:VOLT?"))


class UltraflexInductionHeater(Instrument):
    """UltraFlex Ultraheat S2 (or similar) 2 kW induction heater"""

    name = "UltraflexInductionHeater"

    supported_adapters = (Serial, {})

    knobs = ("power", "output", "mode")

    meters = (
        "voltage",
        "current",
        "frequency",
    )

    @measurer
    def measure_current(self):
        return 0.1 * float(self.query("S3i11K")[3:-3], 16)

    @measurer
    def measure_voltage(self):
        return float(self.query("S3v04K")[3:-3], 16)

    @measurer
    def measure_frequency(self):
        return 1e3 * float(self.query("S3f14K")[3:-3], 16)

    @setter
    def set_power(self, percent_of_max):
        if percent_of_max < 0 or percent_of_max > 100:
            raise ValueError(
                "power setting must be a percentage value between 0 and 100"
            )

        ascii_command = "S34"
        ascii_command += hex(int(round(percent_of_max)))[-2:].upper()
        ascii_command += hex(512 - sum(ascii_command.encode()))[-2:].upper()
        self.query(ascii_command + "K")

    @getter
    def get_power(self):
        return float(self.query("S3B38K")[3:5], 16)

    @setter
    def set_output(self, output):
        if Toggle(output):
            self.query("S3200E8K")
        else:
            self.query("S3347K")

    @setter
    def set_mode(self, mode):
        if mode.lower() == "manual":
            self.query("S3L01CDK")
        elif mode.lower() == "remote":
            self.query("S3L00CEK")
        else:
            raise ValueError(
                f"{self.name}: Unsupported control mode {mode}. "
                'Allowed values are "manual" and "remote".'
            )


class SRSPS300(Instrument):
    """
    Stanford Rsearch Systems PS-300 series high voltage power supply.
    """

    name = "SRSPS350"

    supported_adapters = ((GPIB, {}),)

    knobs = ("output", "voltage", "max voltage" "max current", "trip current")

    presets = {"output": "OFF", "voltage": 0, "max current": 5e-3}

    postsets = {"output": "OFF", "voltage": 0, "max current": 5e-3}

    meters = ("voltage", "current")

    @setter
    def set_output(self, output):
        if Toggle(output):
            self.write("HVON")
        else:
            self.write("HVOF")

    @getter
    def get_output(self):
        # last bit of status byte is the output state

        def validator(response):
            return re.match("\d{1}", response)

        status_bit_7 = int(self.query("*STB? 7", validator=validator))

        if status_bit_7 == 1:
            return "ON"
        else:
            return "OFF"

    @setter
    def set_voltage(self, voltage):
        self.write("VSET%f" % float(voltage))

    @getter
    def get_voltage(self):
        return float(self.query("VSET?"))

    @setter
    def set_max_voltage(self, voltage):
        self.write("VLIM%f" % float(voltage))

    @getter
    def get_max_voltage(self):
        return float(self.query("VLIM?"))

    @setter
    def set_max_current(self, current):
        self.write("ILIM%f" % float(current))

    @getter
    def get_max_current(self):
        return float(self.query("ILIM?"))

    @setter
    def set_trip_current(self, current):
        self.write("ITRP%f" % float(current))

    @getter
    def get_trip_current(self):
        return float(self.query("ITRP?"))

    @measurer
    def measure_voltage(self):
        return float(self.query("VOUT?"))

    @measurer
    def measure_current(self):
        return float(self.query("IOUT?"))



# This module pulls all supported instruments into the same namespace

from empyric.collection.instrument import *
from empyric.collection.sourcemeters import *
from empyric.collection.supplies import *
from empyric.collection.barometers import *
from empyric.collection.thermometers import *
from empyric.collection.multimeters import *
from empyric.collection.controllers import *
from empyric.collection.humans import *
from empyric.collection.spectrometers import *
from empyric.collection.virtual import *
from empyric.collection.generators import *
from empyric.collection.scopes import *
from empyric.collection.io import *

supported = {
    key: value
    for key, value in vars().items()
    if type(value) is type and issubclass(value, Instrument)
}



# Experiment variables

import numbers
import socket
import time
import typing
from functools import wraps
import numpy as np  # functions used in Expression's eval call
from empyric.collection.instrument import Instrument

from empyric import instruments, types
from empyric.tools import write_to_socket, read_from_socket
from empyric.types import *


class Variable:
    """
    Base class representing a specific quantity of interest monitored in an
    experiment.
    """

    type = None  #: the data type of the variable

    #: time since the epoch of last evaluation in seconds, being equal to the
    #: result of `time.time()` being called upon the most recent evaluation of
    #: the `value` property
    last_evaluation = None

    _settable = False  #: whether the variable can be set by the user
    _value = None  #: last known value of the variable

    @property
    def settable(self):
        return self._settable

    @property
    def value(self):
        """The value of the variable"""
        # overwritten by child classes
        pass

    @value.setter
    def value(self, value):
        # overwritten by child classes
        pass

    @staticmethod
    def setter_type_validator(setter):
        """Checks that set value is compatible with variable's type"""

        @wraps(setter)
        def wrapped_setter(self, value):
            if not isinstance(value, Array) and (
                value is None or value == float("nan")
            ):
                self._value = None

            elif self.type is not None:
                setter(self, recast(value, to=self.type))
            else:
                # if type is not explicitly defined upon construction,
                # infer from first set value

                recasted_value = recast(value)

                for _type in types.supported.values():
                    if isinstance(recasted_value, _type):
                        self.type = _type
                        setter(self, recasted_value)

        return wrapped_setter

    @staticmethod
    def getter_type_validator(getter):
        """Checks that get value is compatible with variable's type"""

        @wraps(getter)
        def wrapped_getter(self):
            value = getter(self)

            if not isinstance(value, Array) and (
                value is None or value == float("nan")
            ):
                self._value = None

            elif self.type is not None:
                self._value = recast(value, to=self.type)
            else:
                # if type is not explicitly defined upon construction,
                # infer from first set value

                recasted_value = recast(value)

                for _type in types.supported.values():
                    if isinstance(recasted_value, _type):
                        self.type = _type
                        self._value = recasted_value

            return self._value

        return wrapped_getter


class Knob(Variable):
    """
    Variable that can be directly controlled by an instrument, such as the
    voltage of a power supply.

    The `instrument` argument specifies which instrument the knob is
    associated with.

    The `knob` argument is the label of the knob on the instrument.
    """

    _settable = True  #:

    def __init__(
        self,
        instrument: Instrument,
        knob: str,
        lower_limit: numbers.Number = None,
        upper_limit: numbers.Number = None,
    ):
        self.instrument = instrument
        self.knob = knob  # name of the knob on instrument
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

        # infer type from type hint of first argument of set method
        set_method = getattr(instrument, "set_" + knob.replace(" ", "_"))
        type_hints = typing.get_type_hints(set_method)
        type_hints.pop("return", None)  # exclude return type hint

        if type_hints:
            arg_hints = list(type_hints)
            self._type = type_hints[arg_hints[0]]

        self._value = None

    @property
    @Variable.getter_type_validator
    def value(self):
        """
        Value of the knob of an instrument
        """

        self._value = self.instrument.get(self.knob)
        self.last_evaluation = time.time()

        return self._value

    @value.setter
    @Variable.setter_type_validator
    def value(self, value):
        """
        Set an instrument knob to value
        """

        if self.upper_limit and value > self.upper_limit:
            self.instrument.set(self.knob, self.upper_limit)
        elif self.lower_limit and value < self.lower_limit:
            self.instrument.set(self.knob, self.lower_limit)
        else:
            self.instrument.set(self.knob, value)

        self._value = self.instrument.__getattribute__(self.knob.replace(" ", "_"))


class Meter(Variable):
    """
    Variable that is measured by an instrument, such as temperature.

    Some meters can be controlled directly or indirectly through
    an associated (but distinct) knob.

    The `instrument` argument specifies which instrument the meter is
    associated with.

    The `meter` argument is the label of the meter on the instrument.

    The `gate` optional argument is another variable which gates measurements.
    This is useful for situations where indefinitely continuous measurement of
    the meter, as in the usual experiment loop, is not desirable. Generally,
    it should be a variable of integer, boolean or toggle type. When the gate
    variable evaluates to 1/True/On, the meter can be measured. Otherwise,
    attempts to measure the meter will have no effect (`None` is returned).
    """

    _settable = False  #:

    def __init__(self, instrument: Instrument, meter: str, gate=None):
        self.instrument = instrument
        self.meter = meter

        if gate and isinstance(gate, Variable):
            self.gate = gate
        else:
            self.gate = Parameter(ON)

        self._type = typing.get_type_hints(
            getattr(instrument, "measure_" + meter.replace(" ", "_"))
        ).get("return", None)

        self._value = None

    @property
    @Variable.getter_type_validator
    def value(self):
        """
        Measured value of the meter of an instrument
        """

        if not self.gate.value:
            return None

        self._value = self.instrument.measure(self.meter)
        self.last_evaluation = time.time()

        return self._value


class Expression(Variable):
    """
    Variable that is calculated based on other variables of the experiment

    For example, the output power of a power supply could be recorded as an
    expression, where voltage is a knob and current is a meter:
    power = voltage * current.

    The `expression` argument is a string that can be evaluated by the Python
    interpreter, replacing symbols with the values of variables according to
    the `definitions` argument.

    The `definitions` argument is a dictionary mapping symbols in the
    `expression` argument to variables.
    """

    _settable = False  #:

    _functions = {
        "sqrt(": "np.sqrt(",
        "exp(": "np.exp(",
        "sin(": "np.sin(",
        "cos(": "np.cos(",
        "tan(": "np.tan(",
        "sum(": "np.nansum(",
        "mean(": "np.nanmean(",
        "rms(": "np.nanstd(",
        "std(": "np.nanstd(",
        "var(": "np.nanvar(",
        "diff(": "np.diff(",
        "max(": "np.nanmax(",
        "min(": "np.nanmin(",
    }

    def __init__(self, expression: str, definitions: dict = None):
        self.expression = expression
        self.definitions = definitions if definitions is not None else {}

    @property
    @Variable.getter_type_validator
    def value(self):
        """
        Value of the expression
        """

        expression = self.expression

        # carets represent exponents
        expression = expression.replace("^", "**")

        for symbol, variable in self.definitions.items():
            expression = expression.replace(symbol, f"({variable._value})")

        for shorthand, longhand in self._functions.items():
            if shorthand in expression:
                expression = expression.replace(shorthand, longhand)

        try:
            if "None" not in expression and "nan" not in expression:
                self._value = eval(expression)
            else:
                self._value = None
        except BaseException as err:
            print(
                f"Unable to evaluate expression {self.expression} due to " f"error: ",
                err,
            )
            self._value = None

        self.last_evaluation = time.time()

        return self._value


class Remote(Variable):
    """
    Variable controlled by an experiment (running a server routine) on a
    different process or computer.

    The `server` argument is the IP address and port of the server,
    in the form '(ip address)::(port)'.

    The `alias` argument identifies the particular variable on the server to
    link to. For socket servers, this is simply the name of the variable on
    the server. For Modbus servers, this is the starting address of the
    holding or input register for a read/write or readonly variable,
    respectively. For either set of registers, the starting address is 5*(n-1)
    for the nth variable in the `readwrite` or `readonly` list/dictionary of
    variables in the server routine definition.

    The (optional) `protocol` argument indicates which kind of server to connect
    to. Setting `protocol='modbus'` indicates a Modbus server (controlled by a
    ModbusServer routine on the remote process/computer), and any other value
    or no value indicates a socket server (controlled by a SocketServer
    routine).

    The `settable` argument is required for remote variables on a Modbus server,
    and is not used for a socket server. If `settable` is set to True, then
    the variable value is read from the holding registers (`readwrite`
    variables), otherwise the variable value is read from the input registers
    (`readonly` variables).

    """

    type_map = {
        Toggle: "64bit_uint",
        Boolean: "64bit_uint",
        Integer: "64bit_int",
        Float: "64bit_float",
    }

    def __init__(
        self,
        server: str,
        alias: Union[int, str],
        protocol: str = None,
        settable: bool = False,  # needed for modbus protocol
    ):
        self.server = server
        self.alias = alias
        self.protocol = protocol

        if protocol == "modbus":
            self._client = instruments.ModbusClient(server)
            self._settable = settable

        else:
            server_ip, server_port = server.split("::")

            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            self._socket.connect((server_ip, int(server_port)))

            write_to_socket(self._socket, f"{self.alias} settable?")

            response = read_from_socket(self._socket, timeout=60)
            self._settable = response == f"{self.alias} settable"

            # Get type
            write_to_socket(self._socket, f"{self.alias} type?")

            response = read_from_socket(self._socket, timeout=60)

            if response is not None:
                for _type in types.supported:
                    if str(_type) in response.split(alias)[-1]:
                        self._type = types.supported.get(_type, None)
            else:
                self._type = None

    @property
    @Variable.getter_type_validator
    def value(self):
        """
        Value of the remote variable on a server
        """

        if self.protocol == "modbus":
            fcode = 3 if self.settable else 4

            type_int = self._client.read(fcode, self.alias + 4, _type="16bit_int")

            _type = {
                0: Boolean,
                1: Toggle,
                2: Integer,
                3: Float,
            }.get(type_int, None)

            if _type is not None:
                self._value = self._client.read(
                    fcode, self.alias, count=4, _type=self.type_map[_type]
                )

        else:
            write_to_socket(self._socket, f"{self.alias} ?")

            response = read_from_socket(self._socket, timeout=60)

            try:
                if response is None:
                    self._value = None
                elif "Error" in response:
                    raise RuntimeError(response.split("Error: ")[-1])
                else:
                    self._value = recast(response.split(" ")[-1])

            except BaseException as error:
                print(
                    f"Warning: unable to retrieve value of {self.alias} "
                    f'from server at {self.server}; got error "{error}"'
                )

        return self._value

    @value.setter
    @Variable.setter_type_validator
    def value(self, value):
        """
        Set the value of a remote variable
        """

        if self.protocol == "modbus":
            self._client.write(16, self.alias, value, _type=self.type_map[self.type])

        else:
            write_to_socket(self._socket, f"{self.alias} {value}")

            check = read_from_socket(self._socket, timeout=60)

            if check == "" or check is None:
                print(
                    f"Warning: received no response from server at "
                    f"{self.server} while trying to set {self.alias}"
                )
            elif "Error" in check:
                print(
                    f'Warning: got response "{check}" while trying to set '
                    f"{self.alias} on server at {self.server}"
                )
            else:
                try:
                    check_value = recast(check.split(f"{self.alias} ")[1])

                    if value != check_value:
                        print(
                            f"Warning: attempted to set {self.alias} on "
                            f"server at {self.server} to {value} but "
                            f"checked value is {check_value}"
                        )

                except ValueError as val_err:
                    print(
                        f"Warning: unable to check value while setting "
                        f"{self.alias} on server at {self.server}; "
                        f'got error "{val_err}"'
                    )
                except IndexError as ind_err:
                    print(
                        f"Warning: unable to check value while setting "
                        f"{self.alias} on server at {self.server}; "
                        f'got error "{ind_err}"'
                    )

    def __del__(self):
        if self.protocol == "modbus":
            self._client.disconnect()
        else:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()


class Parameter(Variable):
    """
    Variable whose value is assigned directly by the user or indirectly with a
    routine. An example is a unit conversion factor such as 2.54 cm per inch, a
    numerical constant like pi or a setpoint for a control routine.

    The `parameter` argument is the given value of the parameter.
    """

    _settable = True  #:

    def __init__(self, parameter: Type):
        self.parameter = recast(parameter)
        self._value = parameter

    @property
    @Variable.getter_type_validator
    def value(self):
        """Value of the parameter"""
        return self._value

    @value.setter
    @Variable.setter_type_validator
    def value(self, value):
        """Set the parameter value"""
        self.parameter = value
        self._value = value


supported = {
    key: value
    for key, value in vars().items()
    if type(value) is type and issubclass(value, Variable)
}



import time
import numpy as np
from empyric.instruments import Instrument
from empyric.instruments import setter, measurer
from empyric.adapters import Adapter, Modbus
from empyric.types import Boolean, Toggle, Integer, Float, String, Array


class Clock(Instrument):
    """
    Virtual clock for time keeping; works like a standard stopwatch
    """

    name = "Clock"

    supported_adapters = ((Adapter, {}),)

    knobs = ("state",)
    meters = ("time",)

    @property
    def _time(self):
        if self.stop_time:
            elapsed_time = self.stop_time - self.start_time - self.stoppage
        else:
            elapsed_time = time.time() - self.start_time - self.stoppage

        return elapsed_time

    def __init__(self, *args, **kwargs):
        Instrument.__init__(self, *args, **kwargs)

        # Initially stopped
        self.state = "STOP"
        self.start_time = self.stop_time = time.time()

        self.stoppage = 0  # total time during which the clock has been stopped

    @setter
    def set_state(self, state: String):
        """
        Set the clock state:

        * 'START': setting to this state starts the clock, if it is stopped.
        * 'STOP': setting to this state stops the clock, if it is running.
        * 'RESET': setting to this state resets the clock time to zero.
        """

        if state == "START":
            if self.stop_time:
                self.stoppage += time.time() - self.stop_time
                self.stop_time = False

            self.state = "START"

        elif state == "STOP":
            if not self.stop_time:
                self.stop_time = time.time()

            self.state = "START"

        elif state == "RESET":
            self.start_time = time.time()
            self.stoppage = 0

            if self.stop_time:
                self.stop_time = self.start_time

            return self.state  # return to START or STOP state once reset

    @measurer
    def measure_time(self) -> Float:
        """Measure the clock time"""
        return self._time


class Echo(Instrument):
    """
    Virtual instrument with a single knob "input" and single meter "output",
    useful for testing.

    The "output" meter simply returns the value of the "input" knob.
    """

    name = "Echo"

    supported_adapters = ((Adapter, {}),)

    knobs = ("input",)
    presets = {"input": 0}

    meters = ("output",)

    @setter
    def set_input(self, _input: Float):
        pass

    @measurer
    def measure_output(self) -> Float:
        return self.input


class HenonMapper(Instrument):
    """
    Virtual instrument based on the behavior of a 2D Henon Map:

    x_{n+1} = 1 - a * x_n^2 + y_n

    y_{n+1} = b * x_n

    It has two virtual knobs (a,b) and two virtual meters (x,y)
    """

    name = "HenonMapper"

    supported_adapters = ((Adapter, {}),)

    knobs = ("a", "b")
    presets = {"a": 1.4, "b": 0.3}

    meters = ("x", "y")

    a = 1.4
    b = 0.3

    x, y = 0.63, 0.19  # near the unstable fixed point

    @setter
    def set_a(self, value: Float):
        """
        Set the parameter a

        :param value: (float) new value for a
        :return: None
        """
        pass

    @setter
    def set_b(self, value: Float):
        """
        Set the parameter b

        :param value: (float) new value for b
        :return: None
        """
        pass

    @measurer
    def measure_x(self) -> Float:
        """
        Measure the coordinate x.
        Each call triggers a new iteration, with new values set for x and y
        based on the Henon Map. Therefore, each call to ``measure_x`` should
        be followed by a call to ``measure_y``.

        :return: (float) current value of x
        """

        x_new = 1.0 - self.a * self.x**2 + self.y
        y_new = self.b * self.x

        self.x = x_new
        self.y = y_new

        return self.x

    @measurer
    def measure_y(self) -> Float:
        """
        Measure the coordinate y.
        Each call to ``measure_y`` should be preceded by a call to
        ``measure_x``.

        :return: (float) current value of y
        """

        time.sleep(0.25)  # make sure x is evaluated first

        return self.y


class PIDController(Instrument):
    """
    Virtual PID controller

    """

    name = "PIDController"

    supported_adapters = ((Adapter, {}),)

    knobs = ("setpoint", "proportional gain" "derivative time" "integral time", "input")

    presets = {"proportional gain": 1, "derivative time": 12, "integral time": 180}

    meters = ("output",)

    def __init__(self, *args, **kwargs):
        self.setpoint = None

        Instrument.__init_(self, *args, **kwargs)
        self.clock = Clock()

        self.times = np.array([])
        self.setpoints = np.array([])
        self.inputs = np.array([])
        self.outputs = np.array([])

    @setter
    def set_setpoint(self, setpoint: Float):
        """Set the process setpoint"""
        pass

    @setter
    def set_proportional_gain(self, gain: Float):
        """Set the proportional gain"""
        pass

    @setter
    def set_derivative_time(self, _time: Float):
        """Set the derivative time"""
        pass

    @setter
    def set_integral_time(self, _time: Float):
        """Set the integral time"""
        pass

    @setter
    def set_input(self, input: Float):
        """Input the process value"""

        if len(self.times) == 0:
            self.clock.start()

        if len(self.outputs) == len(self.inputs):
            self.times = np.concatenate([self.times, [self.clock.time]])
            self.setpoints = np.concatenate([self.setpoints, [self.setpoint]])
            self.inputs = np.concatenate([self.inputs, [input]])

    @measurer
    def measure_output(self) -> Float:
        """Get the controller output"""
        if self.setpoint is None:
            # Don't output anything unless the setpoint is defined
            return None

        if len(self.outputs) < len(self.inputs):  # if input has been updated
            # Proportional term
            error = self.setpoint - self.input

            # Integral and derivative terms
            if len(self.times) > 1:
                interval = np.argwhere(
                    self.times >= self.times[-1] - self.integral_time
                ).flatten()

                dt = np.diff(self.times[interval])
                errors = (self.setpoints - self.inputs)[interval[1:]]

                integral = np.sum(dt * errors)

                derivative = -(self.inputs[-1] - self.inputs[-2]) / (
                    self.times[-1] - self.times[-2]
                )
            else:
                integral = 0
                derivative = 0

            tD = self.derivative_time
            tI = self.integral_time

            output = self.proportional_gain * (error + tD * derivative + integral / tI)

            self.outputs.append(output)

            return output
        else:
            return self.outputs[-1]


class RandomWalk(Instrument):
    """
    Virtual random walk process for testing controllers

    Dynamics of the process value is determined by the mean, step and affinity
    knobs. The mean is the mean value of the process in steady state, the step
    is the size of the step that the process can take in either direction upon
    each measurement of the process value, and the affinity is the tendancy of
    the process value to return to its mean value at each step.
    """

    name = "RandomWalk"

    supported_adapters = ((Adapter, {}),)

    knobs = ("mean", "step", "affinity")

    meters = ("value",)

    def __init__(self, *args, **kwargs):
        Instrument.__init__(self, *args, **kwargs)

        self.mean = 0.0
        self.step = 1.0
        self.affinity = 0.01
        self.value = self.mean

    @setter
    def set_mean(self, mean: Float):
        pass

    @setter
    def set_step(self, step: Float):
        pass

    @setter
    def set_drift(self, drift: Float):
        pass

    @measurer
    def measure_value(self) -> Float:
        self.value += np.random.choice([-self.step, self.step]) + self.affinity * (
            self.mean - self.value
        )

        return self.value


class SimpleProcess(Instrument):
    """
    Virtual process that mimics the behavior of a heating process
    """

    name = "SimpleProcess"

    supported_adapters = ((Adapter, {}),)

    knobs = ("setpoint", "noise level", "response time")

    presets = {"setpoint": 0.0, "noise level": 0.1, "response time": 10.0}

    meters = ("value",)

    def __init__(self, *args, **kwargs):
        Instrument.__init__(self, *args, **kwargs)

        self._clock = Clock()
        self._clock.set_state("START")
        self._time = self._clock.measure_time()

    @setter
    def set_setpoint(self, setpoint: Float):
        if hasattr(self, "_time"):
            self.measure_value()
            self._clock.set_state("RESET")
        else:
            self._value = setpoint

    @setter
    def set_noise_level(self, noise_level: Float):
        pass

    @setter
    def set_response_time(self, response_time: Float):
        pass

    @measurer
    def measure_value(self) -> Float:
        last_value = self._value
        t = self._clock.measure_time()  # time since last setpoint change
        self._value = self.setpoint + (last_value - self.setpoint) * np.exp(
            -t / self.response_time
        )

        return self._value + self.noise_level * (2 * np.random.rand() - 1)


class ModbusClient(Instrument):
    """
    Counterpart to the ModbusServer routine. Communicates with a ModbusServer
    instance in other experiments to control variables.
    """

    name = "DataClient"

    supported_adapters = ((Modbus, {}),)



import numbers
import os.path
import time
import sys
import numpy as np
import datetime
import tkinter as tk
import pandas as pd
import pandas.errors

from empyric.types import recast, Type, Float
from empyric.routines import SocketServer, ModbusServer

if sys.platform == "darwin":
    import matplotlib

    matplotlib.use("TkAgg")  # works better on macOS

import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
import matplotlib.dates as mdates

from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()


# needed for converting Pandas datetimes for matplotlib


class Plotter:
    """
    Handler for plotting data based on the runcard plotting settings and data
    context.

    Argument must be a pandas DataFrame with a 'time' column and datetime
    indices.

    The optional settings keyword argument is given in the form,
    {...,
    plot_name:
    {'x': x_name, 'y': [y_name1, y_name2,...], 'style': plot_style, ...} ,
    ...}

    where plot_name is the user designated title for the plot, x_name, y_name_1,
    y_name2, etc. are columns in the DataFrame, and plot_style is either 'basic'
    (default), 'log', 'symlog', 'averaged', 'errorbars' or 'parametric'.
    """

    # For using datetimes on x-axis
    date_locator = mdates.AutoDateLocator()
    date_formatter = mdates.ConciseDateFormatter(date_locator)

    def __init__(self, data, settings=None):
        """
        Plot data based on settings

        :param data: (pandas.Dataframe) data to be plotted.
        :param settings: (dict) dictionary of plot settings
        """

        self.data = data
        self.full_data = self.numericize(data)

        self.plotted = []

        if settings:
            self.settings = settings
        else:
            self.settings = {"Plot": {"x": "Time", "y": data.columns}}

        self.plots = {}
        for plot_name in settings:
            self.plots[plot_name] = plt.subplots(
                constrained_layout=True, figsize=(5, 4)
            )

    def save(self, plot_name=None, save_as=None):
        """Save the plots to PNG files in the working directory"""

        if plot_name:
            fig, ax = self.plots[plot_name]
            if save_as:
                fig.savefig(save_as + ".png")
            else:
                fig.savefig(plot_name + ".png")
        else:
            for name, plot in self.plots.items():
                fig, ax = plot
                fig.savefig(name + ".png")

    def close(self, plot_name=None):
        """
        If the plot_name keyword argument is specified, close the corresponding
        plot. Otherwise, close all plots.
        """

        self.save()

        if plot_name:
            fig, _ = self.plots[plot_name]
            plt.close(fig)
        else:
            plt.close("all")

    def plot(self):
        """Plot all plots"""

        # Update full data
        new_indices = np.setdiff1d(self.data.index, self.full_data.index)
        self.full_data = pd.concat(
            [self.full_data, self.numericize(self.data.loc[new_indices])]
        )

        # Make the plots, by name and style

        for name, settings in self.settings.items():
            style = settings.get("style", "basic")

            if style == "basic":
                self._plot_basic(name)
            elif style == "averaged":
                self._plot_basic(name, averaged=True)
            elif style == "errorbars":
                self._plot_basis(name, errorbars=True)
            elif style == "parametric":
                self._plot_parametric(name)
            else:
                raise AttributeError(f"Plotting style '{style}' not recognized!")

    def _plot_basic(self, name, averaged=False, errorbars=False):
        """Make a simple plot, (possibly multiple) y vs. x"""

        fig, ax = self.plots[name]

        x = self.settings[name].get("x", "Time")
        ys = np.array([self.settings[name]["y"]]).flatten()

        not_in_data = np.setdiff1d(np.concatenate([[x], ys]), self.data.columns)
        if not_in_data.size > 0:
            raise AttributeError(
                f'{", ".join(not_in_data)} specified for plotting, '
                "but not in variables!"
            )

        if averaged or errorbars:
            grouped_data = self.full_data.groupby(x)
            averaged_data = grouped_data.mean()
            xdata = averaged_data[x]
            ydata = averaged_data[ys]

            if errorbars:
                ystddata = grouped_data.std()[ys]

        else:
            if x == "Time":
                xdata = self.full_data.index
            else:
                xdata = self.full_data[x].astype(float)

            ydata = self.full_data[ys].astype(float)

        if ax.lines and not errorbars:
            # For simple plots (i.e. not error bar plots), if already plotted,
            # simply update the plot data and axes

            for line, y in zip(ax.lines, ys):
                line.set_data(xdata, ydata[y])
                ax.draw_artist(line)

            ax.relim()  # reset plot limits based on data
            ax.autoscale_view()

            fig.canvas.draw_idle()
            fig.canvas.start_event_loop(0.01)

        else:  # draw a new plot
            plot_kwargs = self.settings[name].get("configure", {})
            # kwargs for matplotlib

            plot_kwargs = {
                key: np.array([value]).flatten() for key, value in plot_kwargs.items()
            }

            if len(ys) > 1:  # a legend will be made
                plot_kwargs["label"] = ys

            xscale = self.settings[name].get("x scale", "linear")
            yscale = self.settings[name].get("y scale", "linear")

            if x == "Time":
                ax.xaxis.set_major_locator(self.date_locator)
                ax.xaxis.set_major_formatter(self.date_formatter)

            for i, y in enumerate(ys):
                plot_kwargs_i = {key: value[i] for key, value in plot_kwargs.items()}

                if errorbars:
                    ax.errorbar(xdata, ydata[y], yerr=ystddata[y], **plot_kwargs_i)
                else:
                    ax.plot(xdata, ydata[y], **plot_kwargs_i)

            if x == "Time":
                pass
                # axis is automatically configured for datetimes; do not modify
            elif xscale == "linear":
                try:
                    ax.ticklabel_format(axis="x", style="sci", scilimits=(-2, 4))
                except AttributeError:
                    pass
            elif type(xscale) == dict:
                for scale, options in xscale.items():
                    ax.set_xscale(scale, **options)
            else:
                ax.set_xscale(xscale)

            if yscale == "linear":
                try:
                    ax.ticklabel_format(axis="y", style="sci", scilimits=(-2, 4))
                except AttributeError:
                    pass
            elif type(yscale) == dict:
                for scale, options in yscale.items():
                    ax.set_yscale(scale, **options)
            else:
                ax.set_yscale(yscale)

            ax.set_title(name)
            ax.tick_params(labelsize="small")
            ax.grid()

            if len(ys) > 1:
                ax.legend()

            if x != "Time":
                ax.set_xlabel(self.settings[name].get("xlabel", x))
            ax.set_ylabel(self.settings[name].get("ylabel", ys[0]))

            plt.pause(0.01)

    def _plot_parametric(self, name):
        """Make a parametric plot of x and y against a third parameter"""

        fig, ax = self.plots[name]

        x = self.settings[name]["x"]
        y = np.array([self.settings[name]["y"]]).flatten()[0]
        s = self.settings[name].get("s", "Time")

        not_in_data = np.setdiff1d([x, y, s], self.data.columns)
        if not_in_data.size > 0:
            raise AttributeError(
                f'{", ".join(not_in_data)} specified for plotting, '
                "but not in variables!"
            )

        xdata = self.full_data[x].values
        ydata = self.full_data[y].values
        sdata = self.full_data[s].values

        if s == "Time":  # Rescale time if values are large
            units = "seconds"
            if np.max(sdata) > 60:
                units = "minutes"
                sdata = sdata / 60
                if np.max(sdata) > 60:
                    units = "hours"
                    sdata = sdata / 60

        s_min, s_max = np.min(sdata), np.max(sdata)
        norm = plt.Normalize(vmin=s_min, vmax=s_max)

        colormap = "viridis"
        plt.rcParams["image.cmap"] = colormap
        cmap = plt.get_cmap("viridis")

        plot_kwargs = self.settings[name].get("configure", {})
        # kwargs for matplotlib

        if not hasattr(fig, "cbar"):  # draw a new plot
            for i in range(len(xdata) - 1):
                ax.plot(
                    xdata[i : i + 2],
                    ydata[i : i + 2],
                    color=cmap(norm(sdata[i])),
                    **plot_kwargs,
                )

            fig.scalarmappable = ScalarMappable(cmap=cmap, norm=norm)
            fig.scalarmappable.set_array(np.linspace(s_min, s_max, 1000))
            fig.cbar = plt.colorbar(fig.scalarmappable, ax=ax)

            xscale = self.settings[name].get("x scale", "linear")
            yscale = self.settings[name].get("y scale", "linear")

            if xscale == "linear":
                ax.ticklabel_format(axis="x", style="sci", scilimits=(-2, 4))
            elif type(xscale) == dict:
                for scale, options in xscale.items():
                    ax.set_xscale(scale, **options)
            else:
                ax.set_xscale(xscale)

            if yscale == "linear":
                ax.ticklabel_format(axis="y", style="sci", scilimits=(-2, 4))
            elif type(yscale) == dict:
                for scale, options in yscale.items():
                    ax.set_yscale(scale, **options)
            else:
                ax.set_yscale(yscale)

            ax.set_title(name)
            ax.grid(True)
            ax.tick_params(labelsize="small")
            ax.set_xlabel(self.settings[name].get("xlabel", x))
            ax.set_ylabel(self.settings[name].get("ylabel", y))

            if s == "Time":
                fig.cbar.ax.set_ylabel("Time " + f" ({units})")
            else:
                fig.cbar.ax.set_ylabel(s)

            fig.cbar.ax.tick_params(labelsize="small")

            try:
                # Optimize plot layout using the figure's layout engine
                fig.get_layout_engine().execute(fig)
            except AttributeError:  # sometimes happens for reasons
                pass
            except ZeroDivisionError:  # sometimes happens for reasons
                pass

            plt.pause(0.01)

        else:  # if already plotted, update the plot data and axes
            # update color bar
            fig.scalarmappable.set_clim(vmin=s_min, vmax=s_max)
            fig.cbar.update_normal(fig.scalarmappable)

            # update_normal method above removes the colorbar's ylabel;
            # redraw it
            if s == "Time":
                fig.cbar.ax.set_ylabel("Time " + f" ({units})")
            else:
                fig.cbar.ax.set_ylabel(s)

            # update colors of existing line segments
            for i, line in enumerate(ax.lines):
                line.set_color(cmap(norm(sdata[i])))
                ax.draw_artist(line)

            # plot new line segments
            for i in range(len(ax.lines) - 1, len(xdata) - 1):
                ax.plot(
                    xdata[i : i + 2],
                    ydata[i : i + 2],
                    color=cmap(norm(sdata[i])),
                    **plot_kwargs,
                )

            ax.relim()
            ax.autoscale_view()

            fig.canvas.draw_idle()
            fig.canvas.start_event_loop(0.01)

    @staticmethod
    def numericize(data):
        """
        Convert a DataFrame, possibly containing lists or paths,
        such that all elements are numeric
        """

        indices = data.index
        labels = data.columns

        data_array = data.values

        numerical_indices = []
        numerical_array = [[]] * len(labels)

        # Iterate through each row and expand any files or lists into columns
        for index, row in zip(indices, data_array):
            columns = []
            max_len = 0  # maximum length of columns
            for i, element in enumerate(row):
                if type(element) == str and os.path.isfile(element):
                    file_read = False
                    attempt = 0
                    while not file_read:
                        try:
                            expanded_element = list(
                                pd.read_csv(element)[labels[i]].values
                            )

                            file_read = True
                        except pandas.errors.EmptyDataError:
                            if attempt > 3:
                                expanded_element = [np.nan]
                                break

                            attempt += 1
                            plt.pause(1)

                elif np.ndim(element) == 1:
                    expanded_element = list(element)

                    expanded_element = [
                        value if value is not None else np.nan
                        for value in expanded_element
                    ]

                else:
                    if element is None:
                        element = np.nan

                    expanded_element = [element]

                max_len = np.max([len(expanded_element), max_len])

                columns.append(expanded_element)

            numerical_indices = numerical_indices + [index] * max_len
            for i, column in enumerate(columns):
                # fill remainder of column with most recent value
                new_elements = column + [column[-1]] * (max_len - len(column))
                numerical_array[i] = numerical_array[i] + new_elements

        numerical_data = pd.DataFrame(
            data=np.array(numerical_array).T, columns=labels, index=numerical_indices
        )

        return numerical_data


class ExperimentGUI:
    """
    GUI showing experimental progress, values of all experiment variables and
    any alarms, and managing plotting data via the Plotter class.

    This GUI allows the user to hold, stop and terminate the experiment. When
    stopped, the user can change the values of knob and parameter variables,
    and also directly interact with instruments through the Dashboard.
    """

    def __init__(self, experiment, title=None, **kwargs):
        self.experiment = experiment

        if title:
            self.title = title
        else:
            self.title = "Empyric"

        self.closed = False  # has the GUI been closed?

        self.variables = experiment.variables

        self.alarms = kwargs.get("alarms", {})

        if "instruments" in kwargs:
            self.instruments = kwargs["instruments"]
        else:
            # If instruments are not specified,
            # get them from the experiment variables
            self.instruments = {}
            for variable in self.experiment.variables.values():
                if variable.type in ["meter", "knob"]:
                    instrument = variable.instrument
                    if instrument.name not in self.instruments:
                        self.instruments[instrument.name] = instrument

        if ("plots" in kwargs and kwargs["plotter"] is not None) or (
            "plotter" in kwargs and kwargs["plotter"] is not None
        ):
            if "plotter" in kwargs:
                self.plotter = kwargs["plotter"]
            else:
                self.plotter = Plotter(experiment.data, kwargs["plots"])

            self.plot_interval = kwargs.get("plot_interval", 0)
            # grows if plotting takes longer

            self.last_plot = float("-inf")

            # Set interval for saving plots
            self.save_interval = kwargs.get("save_interval", 0)

            self.last_save = time.time()

        self.root = tk.Tk()
        self.root.lift()
        self.root.wm_attributes("-topmost", True)  # bring window to front
        self.root.protocol("WM_DELETE_WINDOW", self.end)

        self.root.title(self.title)
        self.root.resizable(False, False)

        self.status_frame = tk.Frame(self.root)
        self.status_frame.grid(row=0, column=0, columnspan=2)

        i = 0
        title = kwargs.get("title", "Experiment")

        tk.Label(self.status_frame, text=title, font=("Arial", 14, "bold")).grid(
            row=i, column=1
        )

        # Status field shows current experiment status
        i += 1
        tk.Label(
            self.status_frame, text="Status", width=len("Status"), anchor=tk.E
        ).grid(row=i, column=0, sticky=tk.E)

        self.status_label = tk.Label(
            self.status_frame, text="", width=30, relief=tk.SUNKEN
        )
        self.status_label.grid(row=i, column=1, sticky=tk.W, padx=10)

        # Table of variables shows most recently measured/set variable values
        self.variable_entries = {}
        self._entry_enter_funcs = {}  # container for enter press event handlers
        self._entry_esc_funcs = {}  # container for esc press event handlers

        i += 1
        tk.Label(self.status_frame, text="", font=("Arial", 14, "bold")).grid(
            row=i, column=0, sticky=tk.E
        )

        i += 1
        tk.Label(self.status_frame, text="Variables", font=("Arial", 14, "bold")).grid(
            row=i, column=1
        )

        i += 1
        tk.Label(
            self.status_frame, text="Run Time", width=len("Run Time"), anchor=tk.E
        ).grid(row=i, column=0, sticky=tk.E)

        self.variable_entries["Time"] = tk.Entry(
            self.status_frame, state="readonly", disabledforeground="black", width=30
        )
        self.variable_entries["Time"].grid(row=i, column=1, sticky=tk.W, padx=10)

        i += 1
        for name in self.variables:
            tk.Label(self.status_frame, text=name, width=len(name), anchor=tk.E).grid(
                row=i, column=0, sticky=tk.E
            )

            self.variable_entries[name] = tk.Entry(
                self.status_frame,
                state="readonly",
                disabledforeground="black",
                width=30,
            )
            self.variable_entries[name].grid(row=i, column=1, sticky=tk.W, padx=10)

            if self.variables[name].settable:
                entry = self.variable_entries[name]
                variable = self.variables[name]
                root = self.status_frame

                enter_func = lambda event, entry=entry, variable=variable, root=root: self._entry_enter(
                    entry, variable, root
                )

                self._entry_enter_funcs[name] = enter_func
                self.variable_entries[name].bind(
                    "<Return>", self._entry_enter_funcs[name]
                )

                esc_func = lambda event, entry=entry, variable=variable, root=root: self._entry_esc(
                    entry, variable, root
                )

                self._entry_esc_funcs[name] = esc_func
                self.variable_entries[name].bind(
                    "<Escape>", self._entry_esc_funcs[name]
                )

            i += 1

        i += 1
        tk.Label(self.status_frame, text="", font=("Arial", 14, "bold")).grid(
            row=i, column=0, sticky=tk.E
        )

        # Table of alarm indicators shows the status of any alarms monitored
        self.alarm_status_labels = {}

        if len(self.alarms) > 0:
            i += 1
            tk.Label(self.status_frame, text="Alarms", font=("Arial", 14, "bold")).grid(
                row=i, column=1
            )

            i += 1
            for alarm in self.alarms:
                tk.Label(
                    self.status_frame, text=alarm, width=len(alarm), anchor=tk.E
                ).grid(row=i, column=0, sticky=tk.E)

                self.alarm_status_labels[alarm] = tk.Label(
                    self.status_frame, text="Clear", relief=tk.SUNKEN, width=30
                )
                self.alarm_status_labels[alarm].grid(
                    row=i, column=1, sticky=tk.W, padx=10
                )

                i += 1

            tk.Label(self.status_frame, text="", font=("Arial", 14, "bold")).grid(
                row=i, column=0, sticky=tk.E
            )

        # Servers
        self.servers = {
            name: routine
            for name, routine in self.experiment.routines.items()
            if isinstance(routine, SocketServer) or isinstance(routine, ModbusServer)
        }

        if self.servers:
            self.server_status_labels = {}

            i += 1
            tk.Label(
                self.status_frame, text="Servers", font=("Arial", 14, "bold")
            ).grid(row=i, column=1)

            i += 1
            for server in self.servers:
                tk.Label(
                    self.status_frame, text=server, width=len(server), anchor=tk.E
                ).grid(row=i, column=0, sticky=tk.E)

                ip_address = self.servers[server].ip_address
                port = self.servers[server].port

                # Make entry so text is selectable
                self.server_status_labels[server] = tk.Entry(
                    self.status_frame,
                    disabledforeground="black",
                )

                self.server_status_labels[server].insert(0, f"{ip_address}::{port}")
                self.server_status_labels[server].config(state="readonly")

                self.server_status_labels[server].grid(
                    row=i, column=1, sticky=tk.W, padx=10
                )

                i += 1

        # Buttons (root is not self.status_frame, so indexing starts at 1)
        i = 1
        self.dash_button = tk.Button(
            self.root,
            text="Dashboard",
            font=("Arial", 14, "bold"),
            width=10,
            command=self.open_dashboard,
            state=tk.DISABLED,
        )
        self.dash_button.grid(row=i, column=0, sticky=tk.W)

        self.hold_button = tk.Button(
            self.root,
            text="Hold",
            font=("Arial", 14, "bold"),
            width=10,
            command=self.toggle_hold,
        )
        self.hold_button.grid(row=i + 1, column=0, sticky=tk.W)

        self.stop_button = tk.Button(
            self.root,
            text="Stop",
            font=("Arial", 14, "bold"),
            width=10,
            command=self.toggle_stop,
        )
        self.stop_button.grid(row=i + 2, column=0, sticky=tk.W)

        self.terminate_button = tk.Button(
            self.root,
            text="Terminate",
            font=("Arial", 14, "bold"),
            width=10,
            fg="red",
            bg="gray87",
            command=self.end,
            height=5,
        )
        self.terminate_button.grid(row=i, column=1, sticky=tk.E, rowspan=3)

    def run(self):
        """Starts the Tkinter mainloop and the experiment update loop"""
        self.update()
        self.root.mainloop()

    def update(self):
        """Updates the GUI based on the state of the experiment"""

        if self.closed:
            return  # don't update GUI if it no longer exists

        # Allow window to fall back once things get started
        self.root.wm_attributes("-topmost", False)

        state = self.experiment.state

        # Update all variable entries based on experiment state
        for name, entry in self.variable_entries.items():
            # If stopped or holding allow user to edit knobs or parameters
            if self.experiment.stopped or self.experiment.holding:
                if name != "Time" and self.variables[name].settable:
                    continue

            def write_entry(_entry, text):
                _entry.config(state=tk.NORMAL)
                _entry.delete(0, tk.END)
                _entry.insert(0, text)
                _entry.config(state=tk.DISABLED)

            if state[name] is None:
                write_entry(entry, "None")
            elif state[name] == np.nan:
                write_entry(entry, "NaN")
            else:
                if name == "Time":
                    write_entry(entry, str(datetime.timedelta(seconds=state["Time"])))
                elif isinstance(state[name], Float):
                    # display floating point numbers neatly
                    if state[name] == 0.0:
                        write_entry(entry, "0.0")
                    elif np.abs(np.log10(np.abs(state[name]))) > 3:
                        write_entry(entry, "%.3e" % state[name])
                    else:
                        write_entry(entry, "%.3f" % state[name])
                else:
                    write_entry(entry, str(state[name]))

        self.status_label.config(text=self.experiment.status)

        # Check alarms
        for name, label in self.alarm_status_labels.items():
            if self.alarms[name].triggered:
                protocol = self.alarms[name].protocol
                if protocol == "none":
                    label.config(text="TRIGGERED: NO PROTOCOL", bg="red")
                else:
                    label.config(text="TRIGGERED" + f": {protocol.upper()}", bg="red")
            else:
                label.config(text="CLEAR", bg="green")

        # Update hold, stop and dashboard buttons
        if self.experiment.holding or self.experiment.stopped:
            self.dash_button.config(state=tk.NORMAL)

            if self.experiment.holding:
                self.hold_button.config(text="Resume")
                self.stop_button.config(text="Stop")
            else:
                self.hold_button.config(text="Hold")
                self.stop_button.config(text="Resume")

            # Settable variable values can be edited
            for name, entry in self.variable_entries.items():
                if name != "Time" and self.variables[name].settable:
                    entry.config(state=tk.NORMAL)

        else:  # otherwise, experiment is running
            self.dash_button.config(state=tk.DISABLED)
            self.hold_button.config(text="Hold")
            self.stop_button.config(text="Stop")

            for entry in self.variable_entries.values():
                entry.config(state=tk.DISABLED)

        # Quit if experiment has ended
        if self.experiment.terminated:
            self.quit()

        # Plot data
        has_plotter = hasattr(self, "plotter")
        has_data = len(self.experiment.data) > 0
        not_stopped = not self.experiment.stopped
        if has_plotter and has_data and not_stopped:
            if time.time() > self.last_plot + self.plot_interval:
                start_plot = time.perf_counter()
                self.plotter.plot()
                end_plot = time.perf_counter()
                self.last_plot = time.time()

                self.plot_interval = np.max(
                    [self.plot_interval, 5 * int(end_plot - start_plot)]
                )  # adjust interval if drawing plots takes significant time

            # Save plots
            saving_needed = time.time() > self.last_save + self.save_interval
            if saving_needed and self.experiment.timestamp:
                start_save = time.perf_counter()
                self.plotter.save()
                end_save = time.perf_counter()
                self.last_save = time.time()

                self.save_interval = np.max(
                    [self.save_interval, 5 * int(end_save - start_save)]
                )  # adjust interval if saving plots takes significant time

        if not self.closed:
            self.root.after(50, self.update)

    def open_dashboard(self):
        """
        When the user hits the Dashboard button, open a window which allows
        the user to interact with instruments while the experiment is paused or
        stopped.
        """

        prior_status = self.experiment.status

        self.experiment.stop()
        # stop routines and measurements to avoid communication conflicts
        # while dashboard is open

        Dashboard(self.root, self.instruments)

        # Return experiment to prior state
        if "Holding" in prior_status:
            self.experiment.hold()
        elif "Ready" in prior_status or "Running" in prior_status:
            self.experiment.start()

    def toggle_hold(self):
        """User pauses/resumes the experiment through the Hold/Resume button."""

        if self.experiment.holding:
            self.experiment.start()
        else:
            self.experiment.hold()

    def toggle_stop(self):
        """User stops the experiment through the Stop button."""

        if self.experiment.stopped:
            self.experiment.start()
        else:
            self.experiment.stop()

    def end(self):
        """
        User terminates the experiment and closes the GUI with the Terminate
        button; cancels any experiment follow-ups.
        """

        self.experiment.terminate(reason="user terminated")
        self.quit()

    def quit(self):
        if hasattr(self, "plotter"):
            self.plotter.close()

        self.status_label.config(text=self.experiment.TERMINATED)

        self.closed = True
        plt.pause(0.1)  # give GUI and plotter enough time to wrap up
        self.root.update()

        self.root.destroy()
        self.root.quit()

    @staticmethod
    def _entry_enter(entry, variable, root):
        """Assigns the value to a variable if entered by the user"""

        variable.value = recast(
            entry.get(), to=variable.type if variable.type is not None else Type
        )

        root.focus()

    @staticmethod
    def _entry_esc(entry, variable, root):
        """Restores the value of a variable to an entry if user escapes entry"""
        entry.delete(0, tk.END)

        value = variable.value

        try:
            value = float(value)
        except ValueError:
            pass

        if type(value) == float:
            if value == 0:
                entry.insert(0, "0.0")
            elif np.abs(np.log10(np.abs(value))) > 3:
                entry.insert(0, "%.3e" % value)
            else:
                entry.insert(0, "%.3f" % value)
        else:
            entry.insert(0, str(value))

        root.focus()


class BasicDialog(tk.Toplevel):
    """
    General purpose dialog window
    """

    def __init__(self, parent, title=None):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)

        if title:
            self.title(title)

        self.parent = parent

        self.result = None

        body = tk.Frame(self)
        self.initial_focus = self.body(body)
        body.pack(padx=5, pady=5)

        self.buttonbox()

        self.grab_set()

        if not self.initial_focus:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.ok)

        self.geometry("+%d+%d" % (parent.winfo_rootx() + 50, parent.winfo_rooty() + 50))

        self.initial_focus.focus_set()

        self.wait_window(self)

    # construction hooks

    def body(self, master):
        # create dialog body.  return widget that should have
        # initial focus.  this method should be overridden

        pass

    def buttonbox(self):
        # add standard button box. override if you don't want the
        # standard buttons

        box = tk.Frame(self)

        w = tk.Button(box, text="OK", width=10, command=self.ok, default=tk.ACTIVE)
        w.pack(side=tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)

        box.pack()

    # standard button semantics

    def ok(self, event=None):
        if not self.validate():
            self.initial_focus.focus_set()  # put focus back
            return

        self.withdraw()
        self.update_idletasks()

        self.parent.focus_set()
        self.destroy()

    def validate(self):
        return 1  # override


class Dashboard(BasicDialog):
    """
    Allows the user to configure instruments while running up an experiment
    """

    def __init__(self, parent, instruments):
        self.instruments = instruments

        BasicDialog.__init__(self, parent, title="Dashboard")

    def body(self, master):
        tk.Label(master, text="Instruments:", font=("Arial", 14), justify=tk.LEFT).grid(
            row=0, column=0, sticky=tk.W
        )

        i = 1

        self.instrument_labels = {}
        self.config_buttons = {}

        for name, instrument in self.instruments.items():
            instrument_label = tk.Label(master, text=name)
            instrument_label.grid(row=i, column=0)

            self.instrument_labels[name] = instrument_label

            config_button = tk.Button(
                master,
                text="Config/Test",
                command=lambda instr=instrument: self.config(instr),
            )
            config_button.grid(row=i, column=1)

            self.config_buttons[name] = config_button

            i += 1

    def config(self, instrument):
        dialog = ConfigTestDialog(self, instrument)


class ConfigTestDialog(BasicDialog):
    """
    Dialog box for setting knobs and checking meters.
    Allows the user to quickly access basic instrument functionality as well as
    configure instrument for an experiment.
    """

    def __init__(self, parent, instrument):
        self.instrument = instrument
        BasicDialog.__init__(self, parent, title="Config/Test: " + instrument.name)

    def set_knob_entry(self, knob):
        value = self.knob_entries[knob].get()
        self.instrument.set(knob.replace(" ", "_"), recast(value))

    def get_knob_entry(self, knob):
        if hasattr(self.instrument, "get_" + knob.replace(" ", "_")):
            value = self.instrument.get(knob)
        else:
            value = getattr(self.instrument, knob.replace(" ", "_"))

        self.knob_entries[knob].delete(0, tk.END)
        self.knob_entries[knob].insert(0, str(value))

    def update_meter_entry(self, meter):
        value = self.instrument.measure(meter)

        if np.ndim(value) == 1:  # store array data as CSV files
            dataframe = pd.DataFrame({meter: value})
            path = self.instrument.name + "-" + meter.replace(" ", "_") + "_"
            now = datetime.datetime.now()
            path += now.strftime("%Y%m%d-%H%M%S") + ".csv"
            dataframe.to_csv(path)

            value = path

        self.meter_entries[meter].config(state=tk.NORMAL)
        self.meter_entries[meter].delete(0, tk.END)
        self.meter_entries[meter].insert(0, str(value))
        self.meter_entries[meter].config(state="readonly")

    def body(self, master):
        knobs = self.instrument.knobs
        knob_values = {
            knob: getattr(self.instrument, knob.replace(" ", "_")) for knob in knobs
        }
        self.knob_entries = {}

        meters = self.instrument.meters
        self.meter_entries = {}

        label = tk.Label(master, text="Knobs", font=("Arial", 14, "bold"))
        label.grid(row=0, column=0, sticky=tk.W)

        label = tk.Label(master, text="Meters", font=("Arial", 14, "bold"))
        label.grid(row=0, column=3, sticky=tk.W)

        self.set_buttons = {}
        self.get_buttons = {}
        i = 1
        for knob in knobs:
            formatted_name = " ".join(
                [word[0].upper() + word[1:] for word in knob.split(" ")]
            )

            label = tk.Label(master, text=formatted_name)
            label.grid(row=i, column=0, sticky=tk.W)

            self.knob_entries[knob] = tk.Entry(master)
            self.knob_entries[knob].grid(row=i, column=1)
            self.knob_entries[knob].insert(0, str(knob_values[knob]))

            self.set_buttons[knob] = tk.Button(
                master, text="Set", command=lambda knob=knob: self.set_knob_entry(knob)
            )
            self.set_buttons[knob].grid(row=i, column=2)

            self.get_buttons[knob] = tk.Button(
                master, text="Get", command=lambda knob=knob: self.get_knob_entry(knob)
            )
            self.get_buttons[knob].grid(row=i, column=3)

            i += 1

        self.measure_buttons = {}
        i = 1
        for meter in meters:
            formatted_name = " ".join(
                [word[0].upper() + word[1:] for word in meter.split(" ")]
            )

            label = tk.Label(master, text=formatted_name)
            label.grid(row=i, column=4, sticky=tk.W)

            self.meter_entries[meter] = tk.Entry(master)
            self.meter_entries[meter].grid(row=i, column=5)
            self.meter_entries[meter].insert(0, "???")
            self.meter_entries[meter].config(state="readonly")

            self.measure_buttons[meter] = tk.Button(
                master,
                text="Measure",
                command=lambda meter=meter: self.update_meter_entry(meter),
            )
            self.measure_buttons[meter].grid(row=i, column=6)

            i += 1



# Tests for adapters

import sys
import importlib


def test_serial():
    Serial = importlib.import_module("empyric.adapters").Serial

    assert Serial.lib is not None


def test_gpib():
    GPIB = importlib.import_module("empyric.adapters").GPIB

    assert GPIB.lib is not None


def test_usb():
    USB = importlib.import_module("empyric.adapters").USB

    assert USB.lib is not None


def test_modbus():
    Modbus = importlib.import_module("empyric.adapters").Modbus

    assert Modbus.lib is not None


def test_phidget():
    Phidget = importlib.import_module("empyric.adapters").Phidget

    assert Phidget.lib is not None



from empyric.adapters import *
from empyric.types import Toggle, ON, OFF, Float, Integer, String
from empyric.collection.instrument import *


class OmegaCN7500(Instrument):
    """
    Omega model CN7500 PID temperature controller
    """

    name = "OmegaCN7500"

    supported_adapters = (
        (
            ModbusSerial,
            {"slave_mode": "rtu", "baud_rate": 38400, "parity": "N", "delay": 0.2},
        ),
    )

    knobs = (
        "output",
        "setpoint",
        "proportional band",
        "integration time",
        "derivative time",
    )

    meters = ("temperature", "power")

    @setter
    def set_output(self, state: Toggle):
        if state == ON:
            # turn on output & start PID control
            self.backend.write_bit(0x814, 1)
        elif state == OFF:
            # turn off output & stop PID control
            self.backend.write_bit(0x814, 0)

    @setter
    def set_setpoint(self, setpoint: Float):
        self.write(0x1001, int(10 * setpoint))

    @getter
    def get_setpoint(self) -> Float:
        return self.read(0x1001) / 10

    @setter
    def set_proportional_band(self, P: Integer):
        self.write(0x1009, P)

    @getter
    def get_proportional_band(self) -> Integer:
        return self.read(0x1009)

    @setter
    def set_integration_time(self, Ti: Integer):
        self.write(0x100C, Ti)

    @getter
    def get_integration_time(self) -> Integer:
        return self.read(0x100C)

    @setter
    def set_derivative_time(self, Td: Integer):
        self.write(0x100B, Td)

    @getter
    def get_derivative_time(self) -> Integer:
        return self.read(0x100B)

    @measurer
    def measure_temperature(self) -> Float:
        return self.read(0x1000) / 10

    @measurer
    def measure_power(self) -> Float:
        return self.read(0x1000) / 10


class OmegaPlatinum(Instrument):
    """
    Omega Platinum series PID controller
    """

    name = "OmegaPlatinum"

    supported_adapters = ((ModbusSerial, {"baud_rate": 19200, "parity": "O"}),)

    knobs = (
        "setpoint",
        "autotune",
        "output",
        "tc type",
    )

    meters = ("temperature", "power")

    TC_map = {
        "K": 1,
        "J": 0,
        "T": 2,
        "E": 3,
        "N": 4,
        "R": 6,
        "S": 7,
        "B": 8,
        "C": 9,
    }

    inverse_TC_map = {v: k for k, v in TC_map.items()}

    @measurer
    def measure_temperature(self) -> Float:
        return self.read(0x0280, dtype="float")

    @measurer
    def measure_power(self) -> Float:
        return self.read(0x022A, dtype="float")

    @setter
    def set_setpoint(self, setpoint: Float):
        self.write(0x02E2, setpoint, dtype="float")

    @getter
    def get_setpoint(self) -> Float:
        return self.read(0x02E2, dtype="float")

    @setter
    def set_autotune(self, state: Toggle):
        if state == ON:
            self.write(0x0243, 1)
        else:
            self.write(0x0243, 0)

    @setter
    def set_output(self, state: Toggle):
        if state == ON:
            self.write(0x0240, 6)
        else:
            self.write(0x0240, 8)

    @getter
    def get_output(self) -> Toggle:
        state = self.read(0x0240)
        if state == 6 or state == 4:
            return ON
        else:
            return OFF

    @setter
    def set_tc_type(self, _type: String):
        try:
            self.write(0x0283, OmegaPlatinum.TC_map[_type.upper()])
        except (KeyError, AttributeError):
            raise ValueError(f"{self.name}: Invalid thermocouple type {_type}")

    @getter
    def get_tc_type(self) -> String:
        try:
            return self.inverse_TC_map[self.read(0x0283)]
        except KeyError:
            return "None"


class RedLionPXU(Instrument):
    """
    Red Lion PXU temperature PID controller
    """

    name = "RedLionPXU"

    supported_adapters = ((ModbusSerial, {"buad_rate": 38400}),)

    knobs = ("output", "setpoint", "autotune")

    meters = ("temperature", "power")

    @setter
    def set_output(self, state: Toggle):
        if state == ON:
            # turn on output & start PID control
            self.backend.write_bit(0x11, 1)
        elif state == OFF:
            # turn off output & stop PID control
            self.backend.write_bit(0x11, 0)

    @setter
    def set_setpoint(self, setpoint: Integer):
        self.write(0x1, int(setpoint))

    @measurer
    def measure_temperature(self) -> Integer:
        return self.read(0x0)

    @measurer
    def measure_power(self) -> Float:
        return self.read(0x8) / 10

    @setter
    def set_autotune(self, state: Toggle):
        if state == ON:
            self.write(0xF, 1)
        elif state == OFF:
            self.write(0xF, 0)


class WatlowEZZone(Instrument):
    """
    Watlow EZ-Zone PID process controller
    """

    name = "WatlowEZZone"

    supported_adapters = ((ModbusSerial, {"baud_rate": 9600}),)

    knobs = ("setpoint",)

    meters = ("temperature",)

    @measurer
    def measure_temperature(self) -> Float:
        # swapped little-endian byte order (= 3 in minimalmodbus)
        return self.read(360, dtype="float", byte_order=3)

    @getter
    def get_setpoint(self) -> Float:
        return self.read(2160, dtype="float", byte_order=3)

    @setter
    def set_setpoint(self, setpoint: Float):
        return self.write(2160, setpoint, dtype="float", byte_order=3)

    @getter
    def get_proportional_band(self) -> Float:
        return self.read(1890, dtype="float", byte_order=3)

    @setter
    def set_proportional_band(self, band: Float):
        return self.write(1890, band, dtype="float", byte_order=3)

    @getter
    def get_time_integral(self) -> Float:
        return self.read(1894, dtype="float", byte_order=3)

    @setter
    def set_time_integral(self, integral: Float):
        return self.write(1894, integral, dtype="float", byte_order=3)

    @getter
    def get_time_derivative(self) -> Float:
        return self.read(1896, dtype="float", byte_order=3)

    @setter
    def set_time_derivative(self, derivative: Float):
        return self.write(1896, derivative, dtype="float", byte_order=3)


class MKSGSeries(Instrument):
    """MKS G series mass flow controller"""

    name = "MKSGSeries"

    supported_adapters = ((Modbus, {}),)

    knobs = (
        "setpoint",  # flow rate setpoint in SCCM
        "ramp time",  # ramp time in milliseconds
    )

    meters = (
        "flow rate",  # actual flow rate in SCCM
        "valve position",  # valve position in percent
        "temperature",  # temperature in degrees C
    )

    @setter
    def set_setpoint(self, setpoint: Float):
        self.write(16, 0xA000, setpoint, _type="32bit_float")

    @getter
    def get_setpoint(self) -> Float:
        return self.read(3, 0xA000, count=2, _type="32bit_float")

    @setter
    def set_ramp_time(self, ramp_time: Float):
        self.write(16, 0xA002, int(ramp_time), _type="32bit_uint")

    @getter
    def get_ramp_time(self) -> Float:
        return self.read(3, 0xA002, count=2, _type="32bit_uint")

    @measurer
    def measure_flow_rate(self) -> Float:
        return self.read(4, 0x4000, count=2, _type="32bit_float")

    @measurer
    def measure_valve_position(self) -> Float:
        return self.read(4, 0x4004, count=2, _type="32bit_float")

    @measurer
    def measure_temperature(self) -> Float:
        return self.read(4, 0x4002, count=2, _type="32bit_float")


class AlicatMFC(Instrument):
    """Alicat mass flow controller"""

    name = "AlicatMFC"

    supported_adapters = ((Modbus, {}),)

    knobs = ("setpoint",)  # flow rate setpoint in SCCM

    meters = (
        "flow rate",  # actual flow rate in SCCM
        "temperature",  # temperature in degrees C
        "pressure"
        # pressure in PSI (absolute, gauge or differential,
        # depending on device configuration)
    )

    @setter
    def set_setpoint(self, setpoint: Float):
        self.write(16, 1009, setpoint, _type="32bit_float")

    @getter
    def get_setpoint(self) -> Float:
        return self.read(3, 1009, count=2, _type="32bit_float")

    @measurer
    def measure_flow_rate(self) -> Float:
        return self.read(4, 1208, count=2, _type="32bit_float")

    @measurer
    def measure_temperature(self) -> Float:
        return self.read(4, 1204, count=2, _type="32bit_float")

    @measurer
    def measure_pressure(self) -> Float:
        return self.read(4, 1202, count=2, _type="32bit_float")


class SynaccessNetbooter(Instrument):
    supported_adapters = (
        (Socket, {"write_termination": "\r\n", "read_termination": None}),
    )

    knobs = (
        "port 1 toggle",
        "port 2 toggle",
        "port 3 toggle",
        "port 4 toggle",
        "port 5 toggle",
    )

    def _set_port_n_toggle(self, n, state):
        if state != ON and state != OFF:
            raise ValueError(
                f"port toggle state of {self.name} must be "
                f"either ON or OFF (Toggle type)"
            )

        if state == ON:
            self.write("$A3 %d 1" % n)
        elif state == OFF:
            self.write("$A3 %d 0" % n)

    def _get_port_n_toggle(self, n):
        # Dump buffer (this device sends out a Telnet handshake upon initial
        # connection and periodically transmits null bytes, possibly as a
        # keep-alive signal)
        self.read(nbytes=np.inf, timeout=0.1, decode=False)

        def termination(message):
            return re.search(b"A0,\d\d\d\d\d", message)

        status_message = self.query("$A5", termination=termination, decode=False)

        # Port statuses are a sequence of 0s and 1s, starting from the right
        statuses = re.search(b"A0,\d\d\d\d\d", status_message)[0]

        port_n_toggle = ON if int(statuses.decode()[-n]) == 1 else OFF

        return port_n_toggle

    @setter
    def set_port_1_toggle(self, state: Toggle):
        return self._set_port_n_toggle(1, state)

    @getter
    def get_port_1_toggle(self) -> Toggle:
        return self._get_port_n_toggle(1)

    @setter
    def set_port_2_toggle(self, state: Toggle):
        return self._set_port_n_toggle(2, state)

    @getter
    def get_port_2_toggle(self) -> Toggle:
        return self._get_port_n_toggle(2)

    @setter
    def set_port_3_toggle(self, state: Toggle):
        return self._set_port_n_toggle(3, state)

    @getter
    def get_port_3_toggle(self) -> Toggle:
        return self._get_port_n_toggle(3)

    @setter
    def set_port_4_toggle(self, state: Toggle):
        return self._set_port_n_toggle(4, state)

    @getter
    def get_port_4_toggle(self) -> Toggle:
        return self._get_port_n_toggle(4)

    @setter
    def set_port_5_toggle(self, state: Toggle):
        return self._set_port_n_toggle(5, state)

    @getter
    def get_port_5_toggle(self) -> Toggle:
        return self._get_port_n_toggle(5)



import importlib
from empyric.adapters import Phidget, Serial
from empyric.instruments import Instrument, setter, measurer
from empyric.types import recast, String, Float


class Phidget1101(Instrument):
    """
    Phidgets 4x TC reader
    Many instrument methods (setX, getY, etc.) are mapped by the adapter from
    the Phidgets device class
    """

    name = "Phidget1101"

    supported_adapters = ((Phidget, {}),)

    # Available knobs
    knobs = ("type",)

    # Available meters
    meters = ("temperature",)

    def __init__(self, *args, **kwargs):
        self.device_class = importlib.import_module(
            "Phidget22.Devices.TemperatureSensor"
        ).TemperatureSensor

        Instrument.__init__(self, *args, **kwargs)

    @setter
    def set_type(self, type_: String):
        types = importlib.import_module("Phidget22.ThermocoupleType")

        type_dict = {
            "K": types.ThermocoupleType.THERMOCOUPLE_TYPE_K,
            "J": types.ThermocoupleType.THERMOCOUPLE_TYPE_J,
            "T": types.ThermocoupleType.THERMOCOUPLE_TYPE_T,
            "E": types.ThermocoupleType.THERMOCOUPLE_TYPE_E,
        }

        self.write("ThermocoupleType", type_dict[type_])

    @measurer
    def measure_temperature(self) -> Float:
        return self.query("Temperature")


class WilliamsonPyrometer(Instrument):
    """
    Williamson Pro Series 2-color pyrometer.
    """

    supported_adapters = (
        (
            Serial,
            {
                "read_termination": b"\n\r",
                "write_termination": b"\r\n",
                "baud_rate": 38400,
            },
        ),
    )

    meters = (
        "temperature",
        "unfiltered temperature",
        "signal strength",
    )

    @measurer
    def measure_temperature(self) -> Float:
        # temp returned in F, convert to C
        return (recast(self.query("FT")) - 32) / 1.8

    @measurer
    def measure_unfiltered_temperature(self) -> Float:
        # temp returned in F, convert to C
        return (recast(self.query("UT")) - 32) / 1.8

    @measurer
    def measure_signal_strength(self) -> Float:
        return recast(self.query("SS"))



# Tests for the basic features of empyric

import os
import time
import glob
from empyric.variables import Knob, Meter
from empyric.experiment import Experiment, validate_runcard, Manager
from empyric.routines import Timecourse
from empyric.instruments import Echo


def test_experiment(tmp_path):
    """
    Test experiment.Experiment
    """

    os.chdir(str(tmp_path))

    echo = Echo()

    echo_in = Knob(instrument=echo, knob="input")
    echo_out = Meter(instrument=echo, meter="output")

    variables = {"Echo In": echo_in, "Echo Out": echo_out}

    step_up_routine = Timecourse(
        knobs={"Echo In": echo_in},
        times=[0.1, 0.2, 0.3, 0.4, 0.5],
        values=[10, 20, 30, 40, 50],
    )

    routines = {
        "Step Up Timecourse": step_up_routine,
    }

    experiment = Experiment(variables, routines=routines, end="with routines")

    for _ in experiment:
        time.sleep(0.001)

    # check that experiment ended on time
    assert round(experiment.state["Time"], 1) == 0.5
    assert round(echo_out.value) == 50

    # check data saving
    experiment.save()

    assert any(glob.glob("data_*.csv"))


# Use Henon runcard example for testing
tests_dir = os.path.dirname(__file__)

test_runcard_path = os.path.abspath(
    os.path.join(tests_dir, "henon_runcard_example.yaml")
)


def test_runcard_validation():
    try:
        assert os.path.isfile(test_runcard_path)
    except AssertionError:
        raise AssertionError(
            "test runcard 'henon_runcard_example.yaml' was not found in "
            "installed package. This is normally configured by pip running "
            "setup.py. For a proper install, please use pip or similar."
        )

    assert validate_runcard(test_runcard_path)


def test_manager(tmp_path):
    manager = Manager(test_runcard_path)

    # Check that the runcard loaded
    assert manager.description["name"] == "Henon Map Experiment"

    # Run a short version of the Henon Map example experiment
    manager.experiment.end = 10
    manager.followup = None

    manager.run(directory=tmp_path)

    assert manager.experiment.terminated



import re

from empyric.collection.instrument import Instrument, setter, getter, measurer
from empyric.adapters import Socket
from empyric.types import recast, Boolean, Toggle, Integer, Float, Array


class BrainboxesED560(Instrument):
    """
    Brainboxes 4 channel analog output (0-10 V / 0-20mA) gateway.

    For socket communication, the default port is 9500. If the IP address is
    unknown, you can use the Boost.IO Driver software to find it.

    ASCII Protocol must be used.
    """

    supported_adapters = ((Socket, {}),)

    knobs = (
        "analog_out0",
        "analog_out1",
        "analog_out2",
        "analog_out3",
    )

    @setter
    def set_analog_out0(self, value: Float):
        self.query("#010%f" % float(value), validator=self._set_validator)

    @getter
    def get_analog_out0(self) -> Float:
        response = self.query("$0160", validator=self._get_validator)
        return recast(response[3:])

    @setter
    def set_analog_out1(self, value: Float):
        self.query("#011%f" % float(value), validator=self._set_validator)

    @getter
    def get_analog_out1(self) -> Float:
        response = self.query("$0161", validator=self._get_validator)
        return recast(response[3:])

    @setter
    def set_analog_out2(self, value: Float):
        self.query("#012%f" % float(value), validator=self._set_validator)

    @getter
    def get_analog_out2(self) -> Float:
        response = self.query("$0162", validator=self._get_validator)
        return recast(response[3:])

    @setter
    def set_analog_out3(self, value: Float):
        self.query("#013%f" % float(value), validator=self._set_validator)

    @getter
    def get_analog_out3(self) -> Float:
        response = self.query("$0163", validator=self._get_validator)
        return recast(response[3:])

    @staticmethod
    def _get_validator(response):
        return re.match("!01\+\d\d\.\d\d\d", response)

    @staticmethod
    def _set_validator(response):
        return response.strip() == ">"


class BrainboxesED549(Instrument):
    """
    Brainboxes 4 channel analog input (0-10 V / 0-20mA) gateway.

    For socket communication, the default port is 9500. If the IP address is
    unknown, you can use the Boost.IO Driver software to find it.

    ASCII Protocol must be used.
    """

    supported_adapters = ((Socket, {}),)

    meters = (
        "analog_in0",
        "analog_in1",
        "analog_in2",
        "analog_in3",
    )

    @measurer
    def measure_analog_in0(self) -> Float:
        response = self.query("#010", validator=self._validator)
        return response[1:]

    @measurer
    def measure_analog_in1(self) -> Float:
        response = self.query("#011", validator=self._validator)
        return response[1:]

    @measurer
    def measure_analog_in2(self) -> Float:
        response = self.query("#012", validator=self._validator)
        return response[1:]

    @measurer
    def measure_analog_in3(self) -> Float:
        response = self.query("#013", validator=self._validator)
        return response[1:]

    @staticmethod
    def _validator(response):
        return re.match(">\+?\-?\d\d\.\d\d\d", response)



import typing
from threading import RLock
from functools import wraps
from empyric.adapters import *
from empyric.types import recast, Type


def setter(method):
    """
    Utility function that wraps all set_[knob] methods and records the new
    knob values.

    If the wrapped method returns a value, this value is assigned to the
    corresponding knob attribute of the instrument. This is convenient for
    cases where the set value maps to another value which is more relevant.
    Otherwise, the value of the method's first argument is assigned to the
    attribute.

    :param method: (callable) method to be wrapped

    :return: wrapped method
    """

    knob = "_".join(method.__name__.split("_")[1:])

    type_hints = typing.get_type_hints(method)
    type_hints.pop("self", None)
    type_hints.pop("return", None)

    if type_hints:
        dtype = type_hints[list(type_hints)[0]]
    else:
        dtype = Type

    @wraps(method)
    def wrapped_method(*args, **kwargs):
        self = args[0]
        value = args[1]

        self.lock.acquire()

        try:
            returned_value = method(*args, **kwargs)

            # The knob attribute is set to the returned value of the method, or
            # the value argument if the returned value is None
            if returned_value is not None:
                self.__setattr__(knob, recast(returned_value, to=dtype))
            else:
                self.__setattr__(knob, recast(value, to=dtype))
        finally:

            self.lock.release()

    return wrapped_method


def getter(method):
    """
    Utility function that wraps all get_[knob] methods and records the
    retrieved knob values.

    :param method: (callable) method to be wrapped

    :return: wrapped method
    """

    knob = "_".join(method.__name__.split("_")[1:])

    dtype = typing.get_type_hints(method).get("return", Type)

    @wraps(method)
    def wrapped_method(*args, **kwargs):
        self = args[0]

        self.lock.acquire()

        try:
            value = recast(method(*args, **kwargs), to=dtype)
        except AttributeError as err:
            # catches most errors caused by the adapter returning None
            if "NoneType" in str(err):
                value = None
            else:
                raise AttributeError(err)
        finally:
            self.lock.release()

        self.__setattr__(knob, value)

        return value

    return wrapped_method


def measurer(method):
    """
    Utility function that wraps all measure_[meter] methods and records the
    measured value.

    :param method: (callable) method to be wrapped

    :return: wrapped method
    """

    meter = "_".join(method.__name__.split("_")[1:])

    dtype = typing.get_type_hints(method).get("return", Type)

    @wraps(method)
    def wrapped_method(*args, **kwargs):
        self = args[0]

        self.lock.acquire()

        try:
            value = recast(method(*args, **kwargs), to=dtype)
        except AttributeError as err:
            # catches most errors caused by the adapter returning None
            if "NoneType" in str(err):
                value = None
            else:
                raise AttributeError(err)
        finally:
            self.lock.release()

        self.__setattr__(meter, value)

        return value

    return wrapped_method


class Instrument:
    """
    Basic representation of an instrument, essentially a set of knobs and meters

    Each instrument has the following attributes:

    * ``name``: the name of the instrument. Once connected via  an adapter,
      this is converted to
      ``name + '@' + address``.
    * ``supported adapters``: tuple of 2-element tuples. Each 2-element tuple
      contains an adapter class that the instrument can be used with and a
      dictionary of adapter settings.
    * ``knobs``: tuple of the names of all knobs that can be set on the
      instrument.
    * ``presets``: dictionary of knob settings to apply when the instrument is
      instantiated.
      The keys are the names of the knobs and the values are the knob values.
    * ``postsets``: dictionary of knob settings (same format as ``presets``)
      to apply when the instrument is deleted.
    * ``meters``: tuple of the names of all meters that can be measured on
      this instrument.

    """

    name = "Instrument"

    supported_adapters = ((Adapter, {}),)

    knobs = tuple()

    presets = {}

    postsets = {}

    meters = tuple()

    # This lock is used to prevent commands executed in separate threads from
    # interfering with each other. The lock is acquired in the setter, getter
    # and measurer wrapper functions and then released when the wrapped
    # operation is complete. Using an RLock allows set, get and measure
    # methods to call other such methods without blocking, as long as it
    # happens in the same thread, which is the norm.
    lock = RLock()

    def __init__(
        self, address=None, adapter=None, presets=None, postsets=None, **kwargs
    ):
        """

        :param address: (str/int) address of instrument
        :param adapter: (Adapter) desired adapter to use for communications
        with the instrument
        :param presets: (dict) dictionary of instrument presets of the form
        {..., knob: value, ...}
        :param presets: (dict) dictionary of instrument postsets of the form
        {..., knob: value, ...}
        :param kwargs: (dict) any keyword args for the adapter
        """

        if address:
            self.address = address
        else:
            self.address = None

        adapter_connected = False
        if adapter:
            self.adapter = adapter(self, **kwargs)
        else:
            errors = []
            for _adapter, settings in self.supported_adapters:
                settings.update(kwargs)
                try:
                    self.adapter = _adapter(self, **settings)
                    adapter_connected = True
                    break
                except BaseException as error:
                    msg = (
                        f"in trying {_adapter.__name__} adapter, "
                        f"got {type(error).__name__}: {error}"
                    )
                    errors.append(msg)

            if not adapter_connected:
                message = (
                    f"unable to connect an adapter to "
                    f"instrument {self.name} at address {address}:\n"
                )
                for error in errors:
                    message = message + f"{error}\n"
                raise ConnectionError(message)

        if self.address:
            self.name = self.name + "@" + str(self.address)

        # Get existing knob settings, if possible
        for knob in self.knobs:
            if hasattr(self, "get_" + knob.replace(" ", "_")):
                # retrieves the knob value from the instrument
                self.__getattribute__("get_" + knob.replace(" ", "_"))()
            else:
                # knob value is unknown until it is set
                self.__setattr__(knob.replace(" ", "_"), None)

        # Apply presets
        if presets:
            self.presets = {**self.presets, **presets}

        for knob, value in self.presets.items():
            self.set(knob, value)

        # Get postsets
        if postsets:
            self.postsets = {**self.postsets, **postsets}

    def __repr__(self):
        return self.name

    # map write, read and query methods to the adapter's
    def write(self, *args, **kwargs):
        """
        Alias for the adapter's write method

        :param args: any arguments for the adapter's write method, usually
                     including a command string
        :param kwargs: any arguments for the adapter's write method
        :return: whatever is returned by the adapter's write method
        """
        return self.adapter.write(*args, **kwargs)

    def read(self, *args, **kwargs):
        """
        Alias for the adapter's read method

        :param args: any arguments for the adapter's read method
        :param kwargs: any arguments for the adapter's read method
        :return: whatever is returned by the adapter's read method
        """

        return self.adapter.read(*args, **kwargs)

    def query(self, *args, **kwargs):
        """
        Alias for the adapter's query method, if it has one

        :param args: any arguments for the adapter's query method
        :param kwargs: any arguments for the adapter's query method
        :return: whatever is returned by the adapter's query method
        """

        return self.adapter.query(*args, **kwargs)

    def set(self, knob: str, value):
        """
        Set the value of a knob on the instrument

        :param knob: (string) name of variable to be set
        :param value: (float/string) value of new variable setting
        :return: None
        """

        try:
            set_method = getattr(self, "set_" + knob.replace(" ", "_"))
        except AttributeError as err:
            raise AttributeError(f"{knob} cannot be set on {self.name}\n{err}")

        set_method(value)

    def get(self, knob: str):
        """
        Get the value of a knob on the instrument. If the instrument has a get
        method for the knob, a command will be sent to the instrument to
        retrieve the actual value of the knob. If it does not have a get method
        for the knob, the last known value, stored as an instance attribute,
        will be return (possibly being ``nan`` if no value has yet been set)

        """

        if hasattr(self, "get_" + knob.replace(" ", "_")):
            return getattr(self, "get_" + knob.replace(" ", "_"))()
        else:
            return getattr(self, knob.replace(" ", "_"))

    def measure(self, meter: str):
        """
        Measure the value of a variable associated with this instrument

        :param meter: (string) name of the variable to be measured
        :return: (float/string) measured value of the variable
        """

        try:
            measure_method = self.__getattribute__("measure_" + meter.replace(" ", "_"))
        except AttributeError:
            raise AttributeError(f"{meter} cannot be measured on {self.name}")

        measurement = measure_method()

        return measurement

    def disconnect(self):
        """
        Apply any postsets to the instrument and disconnect the adapter

        :return: None
        """

        if self.adapter.connected:
            for knob, value in self.postsets.items():
                self.set(knob, value)

            self.adapter.disconnect()
        else:
            raise ConnectionError(f"adapter for {self.name} is not connected!")



# Function/signal generators
from typing import Union

from empyric.adapters import Socket
from empyric.instruments import Instrument, setter, getter, measurer
from empyric.types import ON, OFF, Toggle, Integer, String, Float, recast


class SiglentSDG1000(Instrument):
    supported_adapters = (
        (Socket, {"write_termination": "\n", "read_termination": "\n"}),
    )

    knobs = (
        "channel 1 output",
        "channel 2 output",
        "channel 1 load",
        "channel 2 load",
        "channel 1 polarity",
        "channel 2 polarity",
        "channel 1 waveform",
        "channel 2 waveform",
        "channel 1 high level",
        "channel 2 high level",
        "channel 1 low level",
        "channel 2 low level",
        "channel 1 frequency",
        "channel 2 frequency",
        "channel 1 pulse width",
        "channel 2 pulse width",
        "channel 1 pulse delay",
        "channel 2 pulse delay",
        'channel 1 invert',
        'channel 2 invert',
        'equal phase'
    )

    wave_forms = ("SINE", "SQUARE", "RAMP", "PULSE", "NOISE", "ARB", "DC", "PRBS", "IQ")

    def _set_channel_n_output(self, n, output: Toggle, load: String, polarity: String):
        self.write(f"C{n}:OUTP {output},LOAD,{load},PLRT,{polarity}")

    def _get_channel_n_output(self, n) -> Toggle:
        response = self.query("C%d:OUTP?" % n).split(",")

        if response[0] == ("C%d:OUTP ON" % n):
            output = ON
        elif response[0] == ("C%d:OUTP OFF" % n):
            output = OFF
        else:
            raise ValueError(
                f"while trying to get output state of {self.name}, "
                f"got corrupted response {response}"
            )

        load = response[2]

        if response[4] == "NOR":
            polarity = "normal"
        elif response[4] == "INVT":
            polarity = "inverted"
        else:
            raise ValueError(
                f"while trying to get output state of {self.name}, "
                f"got corrupted response {response}"
            )

        return output, load, polarity

    def _set_channel_n_waveform(self, n, **kwargs):
        waveform_dict = self._get_channel_n_waveform(n)

        for key in kwargs:
            if key.upper() not in waveform_dict:
                raise ValueError(
                    f"parameter {key} is not valid for waveform type "
                    f'{kwargs["WVTP"]}'
                )

        parameter_string = f"C{n}:BSWV " + ",".join(
            [f"{key.upper()},{value}" for key, value in kwargs.items()]
        )

        # Changing the waveform parameters can cause the relative phase
        # of the two channels to shift
        self.equal_phase = OFF

        self.write(parameter_string)

    def _get_channel_n_waveform(self, n):
        response = self.query(f"C{n}:BSWV?")

        response = response.split(f"C{n}:BSWV ")[1].split(",")

        keys = response[::2]
        values = response[1::2]

        waveform_dict = {key: value for key, value in zip(keys, values)}

        return waveform_dict

    def _set_channel_n_invert(self, n, state: Toggle):

        if state == ON:
            self.write('C%d:INVT ON' % n)
        elif state == OFF:
            self.write('C%d:INVT OFF' % n)

    def _get_channel_n_invert(self, n):

        response = self.query('C%d:INVT?' % n)

        if 'ON' in response:
            return ON
        if 'OFF' in response:
            return OFF

    # Output
    @setter
    def set_channel_1_output(self, output: Toggle):
        _, load, polarity = self._get_channel_n_output(1)

        self._set_channel_n_output(1, output, load, polarity)

    @getter
    def get_channel_1_output(self) -> Toggle:
        return self._get_channel_n_output(1)[0]

    @setter
    def set_channel_2_output(self, output: Toggle):
        _, load, polarity = self._get_channel_n_output(2)

        self._set_channel_n_output(2, output, load, polarity)

    @getter
    def get_channel_2_output(self) -> Toggle:
        return self._get_channel_n_output(2)[0]

    # Load
    @setter
    def set_channel_1_load(self, load: String):
        output, _, polarity = self._get_channel_n_output(1)

        self._set_channel_n_output(1, output, load, polarity)

    @getter
    def get_channel_1_load(self) -> String:
        return self._get_channel_n_output(1)[1]

    @setter
    def set_channel_2_load(self, load: String):
        output, _, polarity = self._get_channel_n_output(2)

        self._set_channel_n_output(2, output, load, polarity)

    @getter
    def get_channel_2_load(self) -> String:
        return self._get_channel_n_output(2)[1]

    # Polarity
    @setter
    def set_channel_1_polarity(self, polarity: String):
        output, load, _ = self._get_channel_n_output(1)

        self._set_channel_n_output(1, output, load, polarity)

    @getter
    def get_channel_1_polarity(self) -> String:
        return self._get_channel_n_output(1)[1]

    @setter
    def set_channel_2_polarity(self, polarity: String):
        output, load, _ = self._get_channel_n_output(2)

        self._set_channel_n_output(2, output, load, polarity)

    @getter
    def get_channel_2_polarity(self) -> String:
        return self._get_channel_n_output(2)[1]

    # Waveform type
    @setter
    def set_channel_1_waveform(self, waveform: String):
        self._set_channel_n_waveform(1, wvtp=waveform)

    @getter
    def get_channel_1_waveform(self) -> String:
        return self._get_channel_n_waveform(1)["WVTP"]

    @setter
    def set_channel_2_waveform(self, waveform: String):
        self._set_channel_n_waveform(2, wvtp=waveform)

    @getter
    def get_channel_2_waveform(self) -> String:
        return self._get_channel_n_waveform(2)["WVTP"]

    # Waveform high level
    @setter
    def set_channel_1_high_level(self, high_level: Union[Float, String]):
        self._set_channel_n_waveform(1, hlev=f"{high_level}")

    @getter
    def get_channel_1_high_level(self) -> Float:
        high_level_str = self._get_channel_n_waveform(1).get("HLEV", "nan")

        return float(high_level_str.replace("V", ""))

    @setter
    def set_channel_2_high_level(self, high_level: Union[Float, String]):
        self._set_channel_n_waveform(2, hlev=f"{high_level}")

    @getter
    def get_channel_2_high_level(self) -> Float:
        high_level_str = self._get_channel_n_waveform(2).get("HLEV", "nan")

        return float(high_level_str.replace("V", ""))

    # Waveform low level
    @setter
    def set_channel_1_low_level(self, low_level: Union[Float, String]):
        self._set_channel_n_waveform(1, llev=f"{low_level}")

    @getter
    def get_channel_1_low_level(self) -> Float:
        low_level_str = self._get_channel_n_waveform(1).get("LLEV", "nan")

        return float(low_level_str.replace("V", ""))

    @setter
    def set_channel_2_low_level(self, low_level: Union[Float, String]):
        self._set_channel_n_waveform(2, llev=f"{low_level}")

    @getter
    def get_channel_2_low_level(self) -> Float:
        low_level_str = self._get_channel_n_waveform(2).get("LLEV", "nan")

        return float(low_level_str.replace("V", ""))

    # Waveform frequency
    @setter
    def set_channel_1_frequency(self, frequency: Union[Float, String]):
        self._set_channel_n_waveform(1, frq=f"{frequency}")

    @getter
    def get_channel_1_frequency(self) -> Float:
        freq_str = self._get_channel_n_waveform(1).get("FRQ", "nan")

        return float(freq_str.replace("HZ", ""))

    @setter
    def set_channel_2_frequency(self, frequency: Union[Float, String]):
        self._set_channel_n_waveform(2, frq=f"{frequency}")

    @getter
    def get_channel_2_frequency(self) -> Float:
        freq_str = self._get_channel_n_waveform(2).get("FRQ", "nan")

        return float(freq_str.replace("HZ", ""))

    # Pulse width
    @setter
    def set_channel_1_pulse_width(self, width: Union[Float, String]):
        self._set_channel_n_waveform(1, width=f"{width}")

    @getter
    def get_channel_1_pulse_width(self) -> Float:
        width_str = self._get_channel_n_waveform(1).get("WIDTH", "nan")

        return float(width_str.replace("S", ""))

    @setter
    def set_channel_2_pulse_width(self, width: Union[Float, String]):
        self._set_channel_n_waveform(2, width=f"{width}")

    @getter
    def get_channel_2_pulse_width(self) -> Float:
        width_str = self._get_channel_n_waveform(2).get("WIDTH", "nan")

        return float(width_str.replace("S", ""))

    # Pulse delay
    @setter
    def set_channel_1_pulse_delay(self, delay: Union[Float, String]):
        self._set_channel_n_waveform(1, dly=f"{delay}")

    @getter
    def get_channel_1_pulse_delay(self) -> Float:
        delay_str = self._get_channel_n_waveform(1).get("DLY", "nan")

        return float(delay_str.replace("S", ""))

    @setter
    def set_channel_2_pulse_delay(self, delay: Union[Float, String]):
        self._set_channel_n_waveform(2, dly=f"{delay}")

    @getter
    def get_channel_2_pulse_delay(self) -> Float:
        delay_str = self._get_channel_n_waveform(2).get("DLY", "nan")

        return float(delay_str.replace("S", ""))

    # Output inversion
    @setter
    def set_channel_1_invert(self, state: Toggle):
        self._set_channel_n_invert(1, state)

    @getter
    def get_channel_1_invert(self) -> Toggle:
        return self._get_channel_n_invert(1)

    @setter
    def set_channel_2_invert(self, state: Toggle):
        self._set_channel_n_invert(2, state)

    @getter
    def get_channel_2_invert(self) -> Toggle:
        return self._get_channel_n_invert(2)

    # Equalize phase of both channels
    @setter
    def set_equal_phase(self, state: Toggle):

        if state == ON:
            self.write('EQPHASE')



# A human is just another instrument
import time

from empyric.adapters import *
from empyric.collection.instrument import *
from empyric.types import String, Float


class ConsoleUser(Instrument):
    """
    Virtual instrument that queries a human operator via the Python console

    The prompt is what the user is asked; the cooldown is the minimum time
    between input requests
    """

    name = "ConsoleUser"

    supported_adapters = ((Adapter, {}),)

    knobs = (
        "prompt",  # message or query to send to user
        "cooldown",
    )  # minimum time before sending repeat messages

    presets = {"cooldown": 0, "prompt": "?"}

    meters = ("response",)

    response = ""

    last_prompt = ""
    last_sent = float("-inf")

    @setter
    def set_prompt(self, prompt: String):
        pass

    @setter
    def set_cooldown(self, cooldown: Float):
        pass

    @measurer
    def measure_response(self) -> String:
        new_prompt = self.prompt != self.last_prompt

        if time.time() >= self.last_sent + self.cooldown or new_prompt:
            response = input(self.prompt)

            self.last_prompt = self.prompt

        return response



from empyric.instruments import Clock
from empyric.variables import Knob, Meter, Parameter, Expression


def test_variable():
    """
    Test experiment.Variable
    """

    clock = Clock()

    test_knob = Knob(instrument=clock, knob="state")

    test_meter = Meter(instrument=clock, meter="time")

    test_parameter = Parameter(parameter=5)

    test_expression = Expression(
        expression="time + offset",
        definitions={"time": test_meter, "offset": test_parameter},
    )

    assert test_knob.value == "STOP"
    assert test_meter.value == 0
    assert test_parameter.value == 5
    assert test_expression.value == 5
